#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2025 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License along with
#  this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
The flight control tab shows telemetry data and flight settings.
"""

import logging
from enum import Enum

from PyQt6 import uic
from PyQt6.QtCore import Qt, pyqtSignal
from PyQt6.QtWidgets import QMessageBox

import cfclient
from cfclient.ui.widgets.ai import AttitudeIndicator

from cflib.crazyflie.log import LogConfig

from cfclient.utils.config import Config
from cfclient.utils.input import JoystickReader

from cfclient.ui.tab_toolbox import TabToolbox

__author__ = 'Bitcraze AB'
__all__ = ['FlightTab']

logger = logging.getLogger(__name__)

flight_tab_class = uic.loadUiType(cfclient.module_path +
                                  "/ui/tabs/flightTab.ui")[0]

MAX_THRUST = 65536.0

TOOLTIP_ALTITUDE_HOLD = """\
Keeps the Crazyflie at its current altitude.
Thrust control becomes height velocity control. The Crazyflie
uses the barometer for height control and uses body-fixed coordinates."""

TOOLTIP_POSITION_HOLD = """\
Keeps the Crazyflie at its current 3D position. Pitch/Roll/
Thrust control becomes X/Y/Z velocity control. Uses world coordinates."""

TOOLTIP_HEIGHT_HOLD = """\
When activated, keeps the Crazyflie at 40cm above the ground.
Thrust control becomes height velocity control. Requires a height
sensor like the Z-Ranger deck or flow deck. Uses body-fixed coordinates.."""

TOOLTIP_HOVER = """\
When activated, keeps the Crazyflie at 40cm above the ground and tries to
keep the position in X and Y as well. Thrust control becomes height velocity
control. Requires a flow deck. Uses body-fixed coordinates."""


class CommanderAction(Enum):
    TAKE_OFF = 1
    LAND = 2
    UP = 3
    DOWN = 4
    LEFT = 5
    RIGHT = 6
    FORWARD = 7
    BACK = 8


class FlightTab(TabToolbox, flight_tab_class):
    uiSetupReadySignal = pyqtSignal()

    _log_data_signal = pyqtSignal(int, object, object)
    _pose_data_signal = pyqtSignal(object, object)

    _input_updated_signal = pyqtSignal(float, float, float, float)
    _rp_trim_updated_signal = pyqtSignal(float, float)
    _emergency_stop_updated_signal = pyqtSignal(bool)
    _arm_updated_signal = pyqtSignal(bool)
    _assisted_control_updated_signal = pyqtSignal(bool)
    _heighthold_input_updated_signal = pyqtSignal(float, float, float, float)
    _hover_input_updated_signal = pyqtSignal(float, float, float, float)

    _log_error_signal = pyqtSignal(object, str)

    # UI_DATA_UPDATE_FPS = 10

    connectionFinishedSignal = pyqtSignal(str)
    disconnectedSignal = pyqtSignal(str)

    _limiting_updated = pyqtSignal(bool, bool, bool)

    LOG_NAME_THRUST = 'stabilizer.thrust'
    LOG_NAME_MOTOR_1 = 'motor.m1'
    LOG_NAME_MOTOR_2 = 'motor.m2'
    LOG_NAME_MOTOR_3 = 'motor.m3'
    LOG_NAME_MOTOR_4 = 'motor.m4'
    LOG_NAME_CAN_FLY = 'sys.canfly'
    LOG_NAME_SUPERVISOR_INFO = 'supervisor.info'

    def __init__(self, helper):
        super(FlightTab, self).__init__(helper, 'Flight Control')
        self.setupUi(self)

        self.disconnectedSignal.connect(self.disconnected)
        self.connectionFinishedSignal.connect(self.connected)
        # Incomming signals
        self._helper.cf.connected.add_callback(
            self.connectionFinishedSignal.emit)
        self._helper.cf.disconnected.add_callback(self.disconnectedSignal.emit)

        self._input_updated_signal.connect(self.updateInputControl)
        self._helper.inputDeviceReader.input_updated.add_callback(
            self._input_updated_signal.emit)
        self._rp_trim_updated_signal.connect(self.calUpdateFromInput)
        self._helper.inputDeviceReader.rp_trim_updated.add_callback(
            self._rp_trim_updated_signal.emit)
        self._emergency_stop_updated_signal.connect(self.updateEmergencyStop)
        self._helper.inputDeviceReader.emergency_stop_updated.add_callback(
            self._emergency_stop_updated_signal.emit)
        self._arm_updated_signal.connect(lambda: self.updateArm(from_controller=True))
        self._helper.inputDeviceReader.arm_updated.add_callback(self._arm_updated_signal.emit)

        self._helper.inputDeviceReader.heighthold_input_updated.add_callback(
            self._heighthold_input_updated_signal.emit)
        self._heighthold_input_updated_signal.connect(
            self._heighthold_input_updated)
        self._helper.inputDeviceReader.hover_input_updated.add_callback(
            self._hover_input_updated_signal.emit)
        self._hover_input_updated_signal.connect(
            self._hover_input_updated)

        self._helper.inputDeviceReader.assisted_control_updated.add_callback(
            self._assisted_control_updated_signal.emit)

        self._assisted_control_updated_signal.connect(
            self._assisted_control_updated)

        self._pose_data_signal.connect(self._pose_data_received)
        self._log_data_signal.connect(self._log_data_received)

        self._log_error_signal.connect(self._logging_error)

        self._isConnected = False

        # Connect UI signals that are in this tab
        self.flightModeCombo.currentIndexChanged.connect(self.flightmodeChange)
        self.minThrust.valueChanged.connect(self.minMaxThrustChanged)
        self.maxThrust.valueChanged.connect(self.minMaxThrustChanged)
        self.thrustLoweringSlewRateLimit.valueChanged.connect(self.thrustLoweringSlewRateLimitChanged)
        self.slewEnableLimit.valueChanged.connect(self.thrustLoweringSlewRateLimitChanged)
        self.targetCalRoll.valueChanged.connect(self._trim_roll_changed)
        self.targetCalPitch.valueChanged.connect(self._trim_pitch_changed)
        self.maxAngle.valueChanged.connect(self.maxAngleChanged)
        self.maxYawRate.valueChanged.connect(self.maxYawRateChanged)
        self.uiSetupReadySignal.connect(self.uiSetupReady)
        self.isInCrazyFlightmode = False

        # Command Based Flight Control
        self._can_fly_deprecated = 0
        self.commanderTakeOffButton.clicked.connect(lambda: self._flight_command(CommanderAction.TAKE_OFF))
        self.commanderLandButton.clicked.connect(lambda: self._flight_command(CommanderAction.LAND))
        self.commanderLeftButton.clicked.connect(lambda: self._flight_command(CommanderAction.LEFT))
        self.commanderRightButton.clicked.connect(lambda: self._flight_command(CommanderAction.RIGHT))
        self.commanderForwardButton.clicked.connect(lambda: self._flight_command(CommanderAction.FORWARD))
        self.commanderBackButton.clicked.connect(lambda: self._flight_command(CommanderAction.BACK))
        self.commanderUpButton.clicked.connect(lambda: self._flight_command(CommanderAction.UP))
        self.commanderDownButton.clicked.connect(lambda: self._flight_command(CommanderAction.DOWN))
        self._update_flight_commander(False)

        # Supervisor
        self._supervisor_info_bitfield = 0
        self.armButton.clicked.connect(self.updateArm)
        self._update_supervisor_and_arming(False)

        self.uiSetupReady()

        self._helper.cf.param.add_update_callback(group="imu_sensors", cb=self._set_available_sensors)

        self._helper.cf.param.all_updated.add_callback(self._all_params_updated)

        self.logAltHold = None

        self.ai = AttitudeIndicator()
        self.verticalLayout_4.addWidget(self.ai)
        self.splitter.setSizes([1000, 1])

        self.targetCalPitch.setValue(Config().get("trim_pitch"))
        self.targetCalRoll.setValue(Config().get("trim_roll"))

        self._tf_state = 0

        # Connect callbacks for input device limiting of roll/pitch/yaw/thrust
        self._helper.inputDeviceReader.limiting_updated.add_callback(self._limiting_updated.emit)
        self._limiting_updated.connect(self._set_limiting_enabled)

        self._helper.pose_logger.data_received_cb.add_callback(self._pose_data_signal.emit)

    def _set_limiting_enabled(self, rp_limiting_enabled, yaw_limiting_enabled, thrust_limiting_enabled):

        self.targetCalRoll.setEnabled(rp_limiting_enabled)
        self.targetCalPitch.setEnabled(rp_limiting_enabled)

        advanced_is_enabled = self.isInCrazyFlightmode
        self.maxAngle.setEnabled(rp_limiting_enabled and advanced_is_enabled)
        self.maxYawRate.setEnabled(yaw_limiting_enabled and advanced_is_enabled)
        self.maxThrust.setEnabled(thrust_limiting_enabled and advanced_is_enabled)
        self.minThrust.setEnabled(thrust_limiting_enabled and advanced_is_enabled)
        self.slewEnableLimit.setEnabled(thrust_limiting_enabled and advanced_is_enabled)
        self.thrustLoweringSlewRateLimit.setEnabled(thrust_limiting_enabled and advanced_is_enabled)

    def thrustToPercentage(self, thrust):
        return ((thrust / MAX_THRUST) * 100.0)

    def uiSetupReady(self):
        flightComboIndex = self.flightModeCombo.findText(Config().get("flightmode"), Qt.MatchFlag.MatchFixedString)
        if (flightComboIndex < 0):
            self.flightModeCombo.setCurrentIndex(0)
            self.flightModeCombo.currentIndexChanged.emit(0)
        else:
            self.flightModeCombo.setCurrentIndex(flightComboIndex)
            self.flightModeCombo.currentIndexChanged.emit(flightComboIndex)

    def _flight_command(self, action):
        current_z = self._helper.pose_logger.position[2]
        move_dist = 0.5
        move_vel = 0.5

        if action == CommanderAction.TAKE_OFF:
            self._helper.cf.param.set_value('commander.enHighLevel', '1')
            z_target = current_z + move_dist
            self._helper.cf.high_level_commander.takeoff(z_target, move_dist / move_vel)
        elif action == CommanderAction.LAND:
            self._helper.cf.high_level_commander.land(0, current_z / move_vel)
        elif action == CommanderAction.LEFT:
            self._helper.cf.high_level_commander.go_to(0, move_dist, 0, 0, move_dist / move_vel, relative=True)
        elif action == CommanderAction.RIGHT:
            self._helper.cf.high_level_commander.go_to(0, -move_dist, 0, 0, move_dist / move_vel, relative=True)
        elif action == CommanderAction.FORWARD:
            self._helper.cf.high_level_commander.go_to(move_dist, 0, 0, 0, move_dist / move_vel, relative=True)
        elif action == CommanderAction.BACK:
            self._helper.cf.high_level_commander.go_to(-move_dist, 0, 0, 0, move_dist / move_vel, relative=True)
        elif action == CommanderAction.UP:
            self._helper.cf.high_level_commander.go_to(0, 0, move_dist, 0, move_dist / move_vel, relative=True)
        elif action == CommanderAction.DOWN:
            self._helper.cf.high_level_commander.go_to(0, 0, -move_dist, 0, move_dist / move_vel, relative=True)

    def _logging_error(self, log_conf, msg):
        QMessageBox.about(self, "Log error",
                          "Error when starting log config [%s]: %s" % (
                              log_conf.name, msg))

    def _log_data_received(self, timestamp, data, logconf):
        if self.isVisible() and self._isConnected:
            self.actualM1.setValue(data[self.LOG_NAME_MOTOR_1])
            self.actualM2.setValue(data[self.LOG_NAME_MOTOR_2])
            self.actualM3.setValue(data[self.LOG_NAME_MOTOR_3])
            self.actualM4.setValue(data[self.LOG_NAME_MOTOR_4])

            self.estimateThrust.setText(
                "%.2f%%" % self.thrustToPercentage(data[self.LOG_NAME_THRUST]))

            if data[self.LOG_NAME_CAN_FLY] != self._can_fly_deprecated:
                self._can_fly_deprecated = data[self.LOG_NAME_CAN_FLY]
                self._update_flight_commander(True)

            if self.LOG_NAME_SUPERVISOR_INFO in data:
                self._supervisor_info_bitfield = data[self.LOG_NAME_SUPERVISOR_INFO]

            self._update_supervisor_and_arming(True)

    def _pose_data_received(self, pose_logger, pose):
        if self.isVisible():
            estimated_z = pose[2]
            roll = pose[3]
            pitch = pose[4]

            self.estimateX.setText(("%.2f" % pose[0]))
            self.estimateY.setText(("%.2f" % pose[1]))
            self.estimateZ.setText(("%.2f" % estimated_z))
            self.estimateRoll.setText(("%.2f" % roll))
            self.estimatePitch.setText(("%.2f" % pitch))
            self.estimateYaw.setText(("%.2f" % pose[5]))

            self.ai.setBaro(estimated_z, self.is_visible())
            self.ai.setRollPitch(-roll, pitch, self.is_visible())

    def _heighthold_input_updated(self, roll, pitch, yaw, height):
        if (self.isVisible() and
                (self._helper.inputDeviceReader.get_assisted_control() ==
                 self._helper.inputDeviceReader.ASSISTED_CONTROL_HEIGHTHOLD)):

            self.targetRoll.setText(("%0.2f deg" % roll))
            self.targetPitch.setText(("%0.2f deg" % pitch))
            self.targetYaw.setText(("%0.2f deg/s" % yaw))
            self.targetHeight.setText(("%.2f m" % height))
            self.ai.setHover(height, self.is_visible())

            self._change_input_labels(using_hover_assist=False)

    def _hover_input_updated(self, vx, vy, yaw, height):
        if (self.isVisible() and
                (self._helper.inputDeviceReader.get_assisted_control() ==
                 self._helper.inputDeviceReader.ASSISTED_CONTROL_HOVER)):

            self.targetRoll.setText(("%0.2f m/s" % vy))
            self.targetPitch.setText(("%0.2f m/s" % vx))
            self.targetYaw.setText(("%0.2f deg/s" % yaw))
            self.targetHeight.setText(("%.2f m" % height))
            self.ai.setHover(height, self.is_visible())

            self._change_input_labels(using_hover_assist=True)

    def _change_input_labels(self, using_hover_assist):
        if using_hover_assist:
            pitch, roll, yaw = 'Velocity X', 'Velocity Y', 'Velocity Z'
        else:
            pitch, roll, yaw = 'Pitch', 'Roll', 'Yaw'

        self.inputPitchLabel.setText(pitch)
        self.inputRollLabel.setText(roll)
        self.inputYawLabel.setText(yaw)

    def _update_supervisor_and_arming(self, connected):
        if not connected:
            self.armButton.setStyleSheet("")
            self.armButton.setText("Arm")
            self.armButton.setEnabled(False)
            self._supervisor_state.setText("")
            self._supervisor_state.setStyleSheet("")
            return

        self._supervisor_state.setText("")
        if self._is_tumbled():
            self._supervisor_state.setText("Tumbled")

        if self._is_locked():
            self.armButton.setText("")
            self.armButton.setEnabled(False)
            self.armButton.setStyleSheet("")
            self._supervisor_state.setText("Locked-please reboot")
            self._supervisor_state.setStyleSheet("background-color: red")
            return
        else:
            self._supervisor_state.setStyleSheet("")

        if self._is_crashed():
            self.armButton.setText("Recover")
            if self._is_tumbled():
                self.armButton.setEnabled(False)
                self.armButton.setStyleSheet("")
                self._supervisor_state.setText("Crashed, flip over to recover")
            else:
                self.armButton.setEnabled(True)
                self.armButton.setStyleSheet("background-color: red")
                self._supervisor_state.setText("Crashed, click Recover")

            self._supervisor_state.setStyleSheet("background-color: red")
            return

        if self._is_flying():
            self.armButton.setEnabled(True)
            self.armButton.setText("Emergency stop")
            self.armButton.setStyleSheet("background-color: red")
            self._supervisor_state.setText("Flying")
            return

        if self._is_armed():
            self.armButton.setStyleSheet("background-color: red")
            if self._auto_arming():
                self.armButton.setEnabled(False)
                self.armButton.setText("Auto armed")
            else:
                self.armButton.setEnabled(True)
                self.armButton.setText("Disarm")
        else:
            self.armButton.setText("Arm")
            if self._can_arm():
                self.armButton.setEnabled(True)
                self.armButton.setStyleSheet("background-color: lightgreen")
            else:
                self.armButton.setStyleSheet("")
                self.armButton.setEnabled(False)

    def _update_flight_commander(self, connected):
        self.commanderBox.setToolTip(str())
        if not connected:
            self.commanderBox.setEnabled(False)
            return

        if self._can_fly_deprecated == 0:
            self.commanderBox.setEnabled(False)
            self.commanderBox.setToolTip('The Crazyflie reports that flight is not possible')
            return

        # We cannot know if we have a positioning deck until we get params
        if not self._helper.cf.param.is_updated:
            self.commanderBox.setEnabled(False)
            return

        #                  flowV1    flowV2     LightHouse       LPS
        position_decks = ['bcFlow', 'bcFlow2', 'bcLighthouse4', 'bcLoco', 'bcDWM1000']
        for deck in position_decks:
            if int(self._helper.cf.param.values['deck'][deck]) == 1:
                self.commanderBox.setEnabled(True)
                break
        else:
            self.commanderBox.setToolTip('You need a positioning deck to use Command Based Flight')
            self.commanderBox.setEnabled(False)
            return

        # To prevent conflicting commands from the controller and the flight panel
        if JoystickReader().available_devices():
            self.commanderBox.setToolTip(
                'Cannot use both a controller and Command Based Flight'
            )
            self.commanderBox.setEnabled(False)
            return

    def connected(self, linkURI):
        self._isConnected = True
        # MOTOR & THRUST
        lg = LogConfig("Motors", Config().get("ui_update_period"))
        lg.add_variable(self.LOG_NAME_THRUST, "uint16_t")
        lg.add_variable(self.LOG_NAME_MOTOR_1)
        lg.add_variable(self.LOG_NAME_MOTOR_2)
        lg.add_variable(self.LOG_NAME_MOTOR_3)
        lg.add_variable(self.LOG_NAME_MOTOR_4)
        lg.add_variable(self.LOG_NAME_CAN_FLY)

        # Add supervisor info if it exists to keep backwards compatibility
        if self._helper.cf.log.toc.get_element_by_complete_name(self.LOG_NAME_SUPERVISOR_INFO):
            lg.add_variable(self.LOG_NAME_SUPERVISOR_INFO)
        # Full supervisor info available after V7, hide supervisor info for earlier versions
        update_supervisor_info = self._helper.cf.platform.get_protocol_version() >= 7
        self._supervisor_state.setVisible(update_supervisor_info)
        self._supervisor_label1.setVisible(update_supervisor_info)
        self._supervisor_label2.setVisible(update_supervisor_info)

        try:
            self._helper.cf.log.add_config(lg)
            lg.data_received_cb.add_callback(self._log_data_signal.emit)
            lg.error_cb.add_callback(self._log_error_signal.emit)
            lg.start()
        except KeyError as e:
            logger.warning(str(e))
        except AttributeError as e:
            logger.warning(str(e))

    def _enable_estimators(self, should_enable):
        self.estimateX.setEnabled(should_enable)
        self.estimateY.setEnabled(should_enable)
        self.estimateZ.setEnabled(should_enable)

    def _set_available_sensors(self, name, available):
        logger.debug("[%s]: %s", name, available)
        available = eval(available)

        self._enable_estimators(True)
        self._helper.inputDeviceReader.set_alt_hold_available(available)

    def disconnected(self, linkURI):
        self._isConnected = False
        self.ai.setRollPitch(0, 0)
        self.actualM1.setValue(0)
        self.actualM2.setValue(0)
        self.actualM3.setValue(0)
        self.actualM4.setValue(0)

        self.estimateRoll.setText("")
        self.estimatePitch.setText("")
        self.estimateYaw.setText("")
        self.estimateThrust.setText("")
        self.estimateX.setText("")
        self.estimateY.setText("")
        self.estimateZ.setText("")

        self.targetHeight.setText("Not Set")
        self.ai.setHover(0, self.is_visible())
        self.targetHeight.setEnabled(False)

        self._enable_estimators(False)

        self.logAltHold = None

        try:
            self._assist_mode_combo.currentIndexChanged.disconnect(
                self._assist_mode_changed)
        except TypeError:
            # Signal was not connected
            pass
        self._assist_mode_combo.setEnabled(False)
        self._assist_mode_combo.clear()

        self._update_flight_commander(False)

        self._supervisor_info_bitfield = 0
        self._update_supervisor_and_arming(False)

    def _can_arm(self):
        return bool(self._supervisor_info_bitfield & 0x0001)

    def _is_armed(self):
        return bool(self._supervisor_info_bitfield & 0x0002)

    def _auto_arming(self):
        return bool(self._supervisor_info_bitfield & 0x0004)

    def _can_fly(self):
        return bool(self._supervisor_info_bitfield & 0x0008)

    def _is_flying(self):
        return bool(self._supervisor_info_bitfield & 0x0010)

    def _is_tumbled(self):
        return bool(self._supervisor_info_bitfield & 0x0020)

    def _is_locked(self):
        return bool(self._supervisor_info_bitfield & 0x0040)

    def _is_crashed(self):
        return bool(self._supervisor_info_bitfield & 0x0080)

    def minMaxThrustChanged(self):
        self._helper.inputDeviceReader.min_thrust = self.minThrust.value()
        self._helper.inputDeviceReader.max_thrust = self.maxThrust.value()
        if (self.isInCrazyFlightmode is True):
            Config().set("min_thrust", self.minThrust.value())
            Config().set("max_thrust", self.maxThrust.value())

    def thrustLoweringSlewRateLimitChanged(self):
        self._helper.inputDeviceReader.thrust_slew_rate = (
            self.thrustLoweringSlewRateLimit.value())
        self._helper.inputDeviceReader.thrust_slew_limit = (
            self.slewEnableLimit.value())
        if (self.isInCrazyFlightmode is True):
            Config().set("slew_limit", self.slewEnableLimit.value())
            Config().set("slew_rate", self.thrustLoweringSlewRateLimit.value())

    def maxYawRateChanged(self):
        logger.debug("MaxYawrate changed to %d", self.maxYawRate.value())
        self._helper.inputDeviceReader.max_yaw_rate = self.maxYawRate.value()
        if (self.isInCrazyFlightmode is True):
            Config().set("max_yaw", self.maxYawRate.value())

    def maxAngleChanged(self):
        logger.debug("MaxAngle changed to %d", self.maxAngle.value())
        self._helper.inputDeviceReader.max_rp_angle = self.maxAngle.value()
        if (self.isInCrazyFlightmode is True):
            Config().set("max_rp", self.maxAngle.value())

    def _trim_pitch_changed(self, value):
        logger.debug("Pitch trim updated to [%f]" % value)
        self._helper.inputDeviceReader.trim_pitch = value
        Config().set("trim_pitch", value)

    def _trim_roll_changed(self, value):
        logger.debug("Roll trim updated to [%f]" % value)
        self._helper.inputDeviceReader.trim_roll = value
        Config().set("trim_roll", value)

    def calUpdateFromInput(self, rollCal, pitchCal):
        logger.debug("Trim changed on joystick: roll=%.2f, pitch=%.2f",
                     rollCal, pitchCal)
        self.targetCalRoll.setValue(rollCal)
        self.targetCalPitch.setValue(pitchCal)

    def updateInputControl(self, roll, pitch, yaw, thrust):
        self.targetRoll.setText(("%0.2f deg" % roll))
        self.targetPitch.setText(("%0.2f deg" % pitch))
        self.targetYaw.setText(("%0.2f deg/s" % yaw))
        self.targetThrust.setText(("%0.2f %%" %
                                   self.thrustToPercentage(thrust)))
        self.thrustProgress.setValue(int(thrust))

        self._change_input_labels(using_hover_assist=False)

    def setMotorLabelsEnabled(self, enabled):
        self.M1label.setEnabled(enabled)
        self.M2label.setEnabled(enabled)
        self.M3label.setEnabled(enabled)
        self.M4label.setEnabled(enabled)

    def emergencyStopStringWithText(self, text):
        return ("<html><head/><body><p>"
                "<span style='font-weight:600; color:#7b0005;'>{}</span>"
                "</p></body></html>".format(text))

    def updateEmergencyStop(self, emergencyStop):
        if emergencyStop:
            self.setMotorLabelsEnabled(False)
            self._helper.cf.loc.send_emergency_stop()
            # TODO krri disarm?
        else:
            self.setMotorLabelsEnabled(True)

    def updateArm(self, from_controller=False):
        if self._is_flying() and not from_controller:
            self._helper.cf.loc.send_emergency_stop()
        elif self._is_crashed():
            self._helper.cf.platform.send_crash_recovery_request()
        elif self._is_armed():
            self._helper.cf.platform.send_arming_request(False)
        elif self._can_arm():
            self.armButton.setStyleSheet("background-color: orange")
            self._helper.cf.platform.send_arming_request(True)

    def flightmodeChange(self, item):
        Config().set("flightmode", str(self.flightModeCombo.itemText(item)))
        logger.debug("Changed flightmode to %s",
                     self.flightModeCombo.itemText(item))
        self.isInCrazyFlightmode = False
        if (item == 0):  # Normal
            self.maxAngle.setValue(Config().get("normal_max_rp"))
            self.maxThrust.setValue(Config().get("normal_max_thrust"))
            self.minThrust.setValue(Config().get("normal_min_thrust"))
            self.slewEnableLimit.setValue(Config().get("normal_slew_limit"))
            self.thrustLoweringSlewRateLimit.setValue(
                Config().get("normal_slew_rate"))
            self.maxYawRate.setValue(Config().get("normal_max_yaw"))
        if (item == 1):  # Advanced
            self.maxAngle.setValue(Config().get("max_rp"))
            self.maxThrust.setValue(Config().get("max_thrust"))
            self.minThrust.setValue(Config().get("min_thrust"))
            self.slewEnableLimit.setValue(Config().get("slew_limit"))
            self.thrustLoweringSlewRateLimit.setValue(
                Config().get("slew_rate"))
            self.maxYawRate.setValue(Config().get("max_yaw"))
            self.isInCrazyFlightmode = True

        if (item == 0):
            newState = False
        else:
            newState = True
        self.maxThrust.setEnabled(newState)
        self.maxAngle.setEnabled(newState)
        self.minThrust.setEnabled(newState)
        self.thrustLoweringSlewRateLimit.setEnabled(newState)
        self.slewEnableLimit.setEnabled(newState)
        self.maxYawRate.setEnabled(newState)

    def _assist_mode_changed(self, item):
        mode = None

        if (item == 0):  # Altitude hold
            mode = JoystickReader.ASSISTED_CONTROL_ALTHOLD
        if (item == 1):  # Position hold
            mode = JoystickReader.ASSISTED_CONTROL_POSHOLD
        if (item == 2):  # Position hold
            mode = JoystickReader.ASSISTED_CONTROL_HEIGHTHOLD
        if (item == 3):  # Position hold
            mode = JoystickReader.ASSISTED_CONTROL_HOVER

        self._helper.inputDeviceReader.set_assisted_control(mode)
        Config().set("assistedControl", mode)

    def _assisted_control_updated(self, enabled):
        if self._helper.inputDeviceReader.get_assisted_control() == \
                JoystickReader.ASSISTED_CONTROL_POSHOLD:
            self.targetThrust.setEnabled(not enabled)
            self.targetRoll.setEnabled(not enabled)
            self.targetPitch.setEnabled(not enabled)
        elif ((self._helper.inputDeviceReader.get_assisted_control() ==
                JoystickReader.ASSISTED_CONTROL_HEIGHTHOLD) or
                (self._helper.inputDeviceReader.get_assisted_control() ==
                 JoystickReader.ASSISTED_CONTROL_HOVER)):
            self.targetThrust.setEnabled(not enabled)
            self.targetHeight.setEnabled(enabled)
            print('Chaning enable for target height: %s' % enabled)
        else:
            self._helper.cf.param.set_value("flightmode.althold", int(enabled))

    def _all_params_updated(self):
        self._populate_assisted_mode_dropdown()
        self._update_flight_commander(True)
        self._update_supervisor_and_arming(True)

    def _populate_assisted_mode_dropdown(self):
        self._assist_mode_combo.addItem("Altitude hold", 0)
        self._assist_mode_combo.addItem("Position hold", 1)
        self._assist_mode_combo.addItem("Height hold", 2)
        self._assist_mode_combo.addItem("Hover", 3)

        # Add the tooltips to the assist-mode items.
        self._assist_mode_combo.setItemData(0, TOOLTIP_ALTITUDE_HOLD, Qt.ItemDataRole.ToolTipRole)
        self._assist_mode_combo.setItemData(1, TOOLTIP_POSITION_HOLD, Qt.ItemDataRole.ToolTipRole)
        self._assist_mode_combo.setItemData(2, TOOLTIP_HEIGHT_HOLD, Qt.ItemDataRole.ToolTipRole)
        self._assist_mode_combo.setItemData(3, TOOLTIP_HOVER, Qt.ItemDataRole.ToolTipRole)

        heightHoldPossible = False
        hoverPossible = False

        if int(self._helper.cf.param.values["deck"]["bcZRanger"]) == 1:
            heightHoldPossible = True
            self._helper.inputDeviceReader.set_hover_max_height(1.0)

        if int(self._helper.cf.param.values["deck"]["bcZRanger2"]) == 1:
            heightHoldPossible = True
            self._helper.inputDeviceReader.set_hover_max_height(2.0)

        if int(self._helper.cf.param.values["deck"]["bcFlow"]) == 1:
            heightHoldPossible = True
            hoverPossible = True
            self._helper.inputDeviceReader.set_hover_max_height(1.0)

        if int(self._helper.cf.param.values["deck"]["bcFlow2"]) == 1:
            heightHoldPossible = True
            hoverPossible = True
            self._helper.inputDeviceReader.set_hover_max_height(2.0)

        if not heightHoldPossible:
            self._assist_mode_combo.model().item(2).setEnabled(False)
        else:
            self._assist_mode_combo.model().item(0).setEnabled(False)

        if not hoverPossible:
            self._assist_mode_combo.model().item(3).setEnabled(False)
        else:
            self._assist_mode_combo.model().item(0).setEnabled(False)

        self._assist_mode_combo.currentIndexChanged.connect(
            self._assist_mode_changed)
        self._assist_mode_combo.setEnabled(True)

        try:
            assistmodeComboIndex = Config().get("assistedControl")
            if assistmodeComboIndex == 3 and not hoverPossible:
                self._assist_mode_combo.setCurrentIndex(0)
                self._assist_mode_combo.currentIndexChanged.emit(0)
            elif assistmodeComboIndex == 0 and hoverPossible:
                self._assist_mode_combo.setCurrentIndex(3)
                self._assist_mode_combo.currentIndexChanged.emit(3)
            elif assistmodeComboIndex == 2 and not heightHoldPossible:
                self._assist_mode_combo.setCurrentIndex(0)
                self._assist_mode_combo.currentIndexChanged.emit(0)
            elif assistmodeComboIndex == 0 and heightHoldPossible:
                self._assist_mode_combo.setCurrentIndex(2)
                self._assist_mode_combo.currentIndexChanged.emit(2)
            else:
                self._assist_mode_combo.setCurrentIndex(assistmodeComboIndex)
                self._assist_mode_combo.currentIndexChanged.emit(
                                                    assistmodeComboIndex)
        except KeyError:
            defaultOption = 0
            if hoverPossible:
                defaultOption = 3
            elif heightHoldPossible:
                defaultOption = 2
            self._assist_mode_combo.setCurrentIndex(defaultOption)
            self._assist_mode_combo.currentIndexChanged.emit(defaultOption)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2023 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License along with
#  this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Shows the Log TOC of available variables in the Crazyflie.
"""

import cfclient
from cfclient.ui.tab_toolbox import TabToolbox
from PyQt6 import QtWidgets
from PyQt6 import uic
from PyQt6.QtCore import pyqtSignal
from PyQt6.QtCore import pyqtSlot
from PyQt6.QtCore import Qt

__author__ = 'Bitcraze AB'
__all__ = ['LogTab']

param_tab_class = uic.loadUiType(cfclient.module_path + "/ui/tabs/logTab.ui")[0]


class LogTab(TabToolbox, param_tab_class):
    connectedSignal = pyqtSignal(str)
    disconnectedSignal = pyqtSignal(str)

    def __init__(self, helper):
        super(LogTab, self).__init__(helper, 'Log TOC')
        self.setupUi(self)

        self.cf = helper.cf

        # Init the tree widget
        self.logTree.setAlternatingRowColors(True)
        if helper.mainUI.isDark:
            self.logTree.setStyleSheet('QTreeWidget { alternate-background-color: #3c3c3c; }')
        else:
            self.logTree.setStyleSheet('QTreeWidget { alternate-background-color: #e9e9e9; }')

        self.logTree.setHeaderLabels(['Name', 'ID', 'Unpack', 'Storage', 'Description'])
        self.logTree.header().resizeSection(0, 150)
        self.logTree.setSortingEnabled(True)
        self.logTree.sortItems(0, Qt.SortOrder.AscendingOrder)

        self.cf.connected.add_callback(self.connectedSignal.emit)
        self.connectedSignal.connect(self.connected)

        # Clear the log TOC list when the Crazyflie is disconnected
        self.cf.disconnected.add_callback(self.disconnectedSignal.emit)
        self.disconnectedSignal.connect(self.disconnected)

    @pyqtSlot('QString')
    def disconnected(self, linkname):
        root = self.logTree.invisibleRootItem()
        for i in range(root.childCount()):
            item = root.child(i)
            item.setFlags(Qt.ItemFlag.NoItemFlags)

    @pyqtSlot(str)
    def connected(self, linkURI):
        self.logTree.clear()

        toc = self.cf.log.toc

        for row_idx, group in enumerate(list(toc.toc.keys())):
            groupItem = QtWidgets.QTreeWidgetItem()
            groupItem.setData(0, Qt.ItemDataRole.DisplayRole, group)

            for param in list(toc.toc[group].keys()):
                item = QtWidgets.QTreeWidgetItem()
                item.setData(0, Qt.ItemDataRole.DisplayRole, param)
                item.setData(1, Qt.ItemDataRole.DisplayRole, toc.toc[group][param].ident)
                item.setData(2, Qt.ItemDataRole.DisplayRole, toc.toc[group][param].pytype)
                item.setData(3, Qt.ItemDataRole.DisplayRole, toc.toc[group][param].ctype)

                if cfclient.log_param_doc is not None:
                    try:
                        log_groups = cfclient.log_param_doc['logs'][group]
                        log_variable = log_groups['variables'][param]
                        item.setData(4, Qt.ItemDataRole.DisplayRole, log_variable['short_desc'])
                    except (KeyError, TypeError, AttributeError):
                        pass

                groupItem.addChild(item)

            self.logTree.addTopLevelItem(groupItem)

            if cfclient.log_param_doc is not None:
                try:
                    log_groups = cfclient.log_param_doc['logs'][group]
                    label = QtWidgets.QLabel(log_groups['desc'])
                    label.setWordWrap(True)
                    self.logTree.setItemWidget(groupItem, 4, label)
                except (KeyError, TypeError, AttributeError):
                    pass



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License along with
#  this program; if not, write to the Free Software Foundation, Inc., 51
#  Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import json
import os
from appdirs import AppDirs
import sys

# Path used all over the application
if not hasattr(sys, 'frozen'):
    module_path = os.path.dirname(__file__)
else:
    module_path = os.path.join(os.path.dirname(sys.executable),
                               'lib', 'cfclient')
config_path = AppDirs("cfclient", "Bitcraze").user_config_dir

if not hasattr(sys, 'frozen'):
    from importlib.metadata import version, PackageNotFoundError
    try:
        VERSION = version("cfclient")
    except PackageNotFoundError:
        VERSION = "dev"
else:
    try:
        from .version import __version__ as imported_version
        VERSION = imported_version
    except ImportError:
        VERSION = "dev"
try:
    with open(os.path.join(module_path, "resources/log_param_doc.json")) as f:
        log_param_doc = json.load(f)
except (IOError, OSError, json.JSONDecodeError):
    log_param_doc = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2023 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License along with
#  this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
Shows all the parameters available in the Crazyflie and also gives the ability
to edit them.
"""

import logging
from threading import Event

from PyQt6 import uic, QtCore
from PyQt6.QtCore import QSortFilterProxyModel, Qt, pyqtSignal
from PyQt6.QtCore import QAbstractItemModel, QModelIndex, QVariant
from PyQt6.QtGui import QBrush, QColor
from PyQt6.QtWidgets import QHeaderView, QFileDialog, QMessageBox

from cflib.crazyflie.param import PersistentParamState
from cflib.localization import ParamFileManager

import cfclient
from cfclient.ui.tab_toolbox import TabToolbox
from cfclient.utils.logconfigreader import FILE_REGEX_YAML

__author__ = 'Bitcraze AB'
__all__ = ['ParamTab']

param_tab_class = uic.loadUiType(cfclient.module_path + "/ui/tabs/paramTab.ui")[0]

logger = logging.getLogger(__name__)


def round_if_float(value):
    """If the value is float, we limit to 5 significat numbers"""
    try:
        value = float(value)
        value = f'{value:.5g}'
    except ValueError:
        pass
    return value


class ParamChildItem(object):
    """Represents a leaf-node in the tree-view (one parameter)"""

    def __init__(self, parent, name, persistent, crazyflie):
        """Initialize the node"""
        self.parent = parent
        self.name = name
        self.ctype = None
        self.access = None
        self.persistent = False
        self.value = ""
        self._cf = crazyflie
        self.is_updating = True
        self.state = None

    def updated(self, name, value):
        """Callback from the param layer when a parameter has been updated"""
        self.value = round_if_float(value)
        self.is_updating = False
        self.parent.model.proxy.dataChanged.emit(QModelIndex(), QModelIndex())

    def child_count(self):
        """Return the number of children this node has"""
        return 0


class ParamGroupItem(object):
    """Represents a parameter group in the tree-view"""

    def __init__(self, name, model):
        """Initialize the parent node"""
        super(ParamGroupItem, self).__init__()
        self.parent = None
        self.children = []
        self.name = name
        self.model = model

    def child_count(self):
        """Return the number of children this node has"""
        return len(self.children)


class ParamBlockModel(QAbstractItemModel):
    """Model for handling the parameters in the tree-view"""

    def __init__(self, parent, mainUI):
        """Create the empty model"""
        super(ParamBlockModel, self).__init__(parent)
        self._nodes = []
        self._column_headers = ['Name', 'Type', 'Access', 'Persistent', 'Value']
        self._red_brush = QBrush(QColor("red"))
        self._enabled = False
        self._mainUI = mainUI
        self.proxy = None

    def set_proxy(self, proxy):
        self.proxy = proxy

    def set_enabled(self, enabled):
        if self._enabled != enabled:
            self._enabled = enabled
            self.layoutChanged.emit()

    def set_toc(self, toc, crazyflie):
        """Populate the model with data from the param TOC"""

        # No luck using proxy sorting, so do it here instead...
        for group in sorted(toc.keys()):
            new_group = ParamGroupItem(group, self)
            for param in sorted(toc[group].keys()):
                elem = toc[group][param]
                is_persistent = elem.is_persistent()
                new_param = ParamChildItem(new_group, param, is_persistent, crazyflie)
                new_param.ctype = elem.ctype
                new_param.access = elem.get_readable_access()
                new_param.persistent = elem.is_persistent()

                crazyflie.param.add_update_callback(
                    group=group, name=param, cb=new_param.updated)
                new_group.children.append(new_param)
            self._nodes.append(new_group)

        self.layoutChanged.emit()

    def refresh(self):
        """Force a refresh of the view though the model"""
        self.layoutChanged.emit()

    def parent(self, index):
        """Re-implemented method to get the parent of the given index"""
        if not index.isValid():
            return QModelIndex()

        node = index.internalPointer()
        if node.parent is None:
            return QModelIndex()
        else:
            return self.createIndex(self._nodes.index(node.parent), 0,
                                    node.parent)

    def columnCount(self, parent):
        """Re-implemented method to get the number of columns"""
        return len(self._column_headers)

    def headerData(self, section, orientation, role):
        """Re-implemented method to get the headers"""
        if role == Qt.ItemDataRole.DisplayRole:
            return self._column_headers[section]

    def rowCount(self, parent):
        """Re-implemented method to get the number of rows for a given index"""
        parent_item = parent.internalPointer()
        if parent.isValid():
            parent_item = parent.internalPointer()
            return parent_item.child_count()
        else:
            return len(self._nodes)

    def index(self, row, column, parent):
        """Re-implemented method to get the index for a specified
        row/column/parent combination"""
        if not self._nodes:
            return QModelIndex()
        node = parent.internalPointer()
        if not node:
            index = self.createIndex(row, column, self._nodes[row])
            self._nodes[row].index = index
            return index
        else:
            return self.createIndex(row, column, node.children[row])

    def data(self, index, role):
        """Re-implemented method to get the data for a given index and role"""

        if role == Qt.ItemDataRole.BackgroundRole:
            if index.row() % 2 == 0:
                return QVariant(self._mainUI.bgColor)
            else:
                multiplier = 1.15 if self._mainUI.isDark else 0.95
                return QVariant(
                    QColor(
                        int(self._mainUI.bgColor.red() * multiplier),
                        int(self._mainUI.bgColor.green() * multiplier),
                        int(self._mainUI.bgColor.blue() * multiplier)
                    )
                )

        node = index.internalPointer()
        parent = node.parent
        if not parent:
            if role == Qt.ItemDataRole.DisplayRole and index.column() == 0:
                return node.name
        elif role == Qt.ItemDataRole.DisplayRole:
            if index.column() == 0:
                return node.name
            if index.column() == 1:
                return node.ctype
            if index.column() == 2:
                return node.access
            if index.column() == 3:
                return 'Yes' if node.persistent else 'No'
            if index.column() == 4:
                return node.value
        elif (role == Qt.ItemDataRole.BackgroundRole and index.column() == 4 and
              node.is_updating):
            return self._red_brush

        return None

    def flags(self, index):
        """Re-implemented function for getting the flags for a certain index"""
        flag = super(ParamBlockModel, self).flags(index)

        if not self._enabled:
            return Qt.ItemFlag.NoItemFlags

        return flag

    def reset(self):
        """Reset the model"""
        super(ParamBlockModel, self).beginResetModel()
        self._nodes = []
        super(ParamBlockModel, self).endResetModel()
        self.layoutChanged.emit()


class ParamTreeFilterProxy(QSortFilterProxyModel):
    """
    Implement a filtering proxy model that show all children if the group matches.
    """
    def __init__(self, paramTree):
        super(ParamTreeFilterProxy, self).__init__(paramTree)

    def filterAcceptsRow(self, source_row: int, source_parent: QModelIndex) -> bool:
        '''
        When a group match the filter, make sure all children matches as well.
        '''
        if not source_parent.isValid():
            return super().filterAcceptsRow(source_row, source_parent)

        return super().filterAcceptsRow(source_parent.row(), source_parent.parent())


class ParamTab(TabToolbox, param_tab_class):
    """
    Show all the parameters in the TOC and give the user the ability to edit
    them
    """
    _expand_all_signal = pyqtSignal()
    _connected_signal = pyqtSignal(str)
    _disconnected_signal = pyqtSignal(str)

    _set_param_value_signal = pyqtSignal()
    _persistent_state_signal = pyqtSignal(PersistentParamState)
    _param_default_signal = pyqtSignal(object)
    _reset_param_signal = pyqtSignal(str)

    def __init__(self, helper):
        """Create the parameter tab"""
        super(ParamTab, self).__init__(helper, 'Parameters')
        self.setupUi(self)

        self.cf = helper.cf

        self.cf.connected.add_callback(self._connected_signal.emit)
        self._connected_signal.connect(self._connected)
        self.cf.disconnected.add_callback(self._disconnected_signal.emit)
        self._disconnected_signal.connect(self._disconnected)

        self._model = ParamBlockModel(None, self._helper.mainUI)
        self._persistent_state_signal.connect(self._persistent_state_cb)
        self._set_param_value_signal.connect(self._set_param_value)
        self.setParamButton.clicked.connect(self._set_param_value_signal.emit)
        self.currentValue.returnPressed.connect(self._set_param_value_signal.emit)
        self._param_default_signal.connect(self._param_default_cb)

        self._reset_param_signal.connect(lambda text: self.currentValue.setText(text))
        self.resetDefaultButton.clicked.connect(lambda: self._reset_param_signal.emit(self.defaultValue.text()))
        self.persistentButton.clicked.connect(self._persistent_button_cb)

        self.proxyModel = ParamTreeFilterProxy(self.paramTree)
        self.proxyModel.setSourceModel(self._model)
        self.proxyModel.setRecursiveFilteringEnabled(True)
        self._model.set_proxy(self.proxyModel)

        @QtCore.pyqtSlot(str)
        def onFilterChanged(text):
            self.proxyModel.setFilterRegExp(text)

        self.filterBox.textChanged.connect(onFilterChanged)

        self.paramTree.setModel(self.proxyModel)
        self.paramTree.header().setSectionResizeMode(QHeaderView.ResizeMode.ResizeToContents)
        self.paramTree.selectionModel().selectionChanged.connect(self._paramChanged)

        self._load_param_button.clicked.connect(self._load_param_button_clicked)
        self._dump_param_button.clicked.connect(self._dump_param_button_clicked)
        self._clear_param_button.clicked.connect(self._clear_stored_persistent_params_button_clicked)

        self._is_connected = False
        self._update_param_io_buttons()

    def _param_default_cb(self, default_value):
        if default_value is not None:
            self.defaultValue.setText(str(default_value))
        else:
            self.defaultValue.setText('-')

    def _persistent_button_cb(self, _):
        def success_cb(name, success):
            print(f'store {success}!')
            if success:
                self.cf.param.persistent_get_state(name, lambda _, state: self._persistent_state_signal.emit(state))

        complete = self.paramDetailsLabel.text()
        if self.persistentButton.text() == 'Clear':
            self.cf.param.persistent_clear(complete, success_cb)
        else:
            self.cf.param.persistent_store(complete, success_cb)

    def _persistent_state_cb(self, state):
        print(f'persistent callback! state: {state}')
        self.persistentFrame.setVisible(True)

        if state.is_stored:
            self.storedValue.setText(str(state.stored_value))
        else:
            self.storedValue.setText('Not stored')

        self.persistentButton.setText('Clear' if state.is_stored else 'Store')

    def _set_param_value(self):
        name = self.paramDetailsLabel.text()
        value = self.currentValue.text()
        try:
            self.cf.param.set_value(name, value)
            self.currentValue.setStyleSheet('')
        except Exception:
            self.currentValue.setStyleSheet('border: 1px solid red')

    def _paramChanged(self):

        group = None
        param = None
        indexes = self.paramTree.selectionModel().selectedIndexes()
        if len(indexes) > 0:
            selectedIndex = indexes[0]
            if selectedIndex.parent().isValid():
                group = selectedIndex.parent().data()
                param = selectedIndex.data()
            else:
                group = selectedIndex.data()

        # Made visible in _persistent_state_cb()
        self.persistentFrame.setVisible(False)

        are_details_visible = param is not None
        self.valueFrame.setVisible(are_details_visible)
        self.paramDetailsLabel.setVisible(are_details_visible)
        self.paramDetailsDescription.setVisible(are_details_visible)

        if param:
            self.paramDetailsLabel.setText(f'{group}.{param}' if param is not None else group)
            if cfclient.log_param_doc is not None:
                try:
                    desc = str()
                    group_doc = cfclient.log_param_doc['params'][group]
                    if param is None:
                        desc = group_doc['desc']
                    else:
                        desc = group_doc['variables'][param]['short_desc']

                    self.paramDetailsDescription.setWordWrap(True)
                    self.paramDetailsDescription.setText(desc.replace('\n', ''))
                except (KeyError, TypeError, AttributeError):
                    self.paramDetailsDescription.setText('')

            complete = f'{group}.{param}'
            elem = self.cf.param.toc.get_element_by_complete_name(complete)
            value = round_if_float(self.cf.param.get_value(complete))
            self.currentValue.setText(value)
            self.currentValue.setStyleSheet('')
            self.currentValue.setCursorPosition(0)
            self.defaultValue.setText('-')
            self.cf.param.get_default_value(complete, lambda _, value: self._param_default_signal.emit(value))

            writable = elem.get_readable_access() == 'RW'
            self.currentValue.setEnabled(writable)
            self.setParamButton.setEnabled(writable)
            self.resetDefaultButton.setEnabled(writable)

            if elem.is_persistent():
                self.cf.param.persistent_get_state(complete, lambda _, state: self._persistent_state_signal.emit(state))

    def _update_param_io_buttons(self):
        enabled = self._is_connected
        self._load_param_button.setEnabled(enabled)
        self._dump_param_button.setEnabled(enabled)
        self._clear_param_button.setEnabled(enabled)

    def _load_param_button_clicked(self):
        names = QFileDialog.getOpenFileName(self, 'Open file', cfclient.config_path, FILE_REGEX_YAML)

        if names[0] == '':
            return
        filename = names[0]
        parameters = ParamFileManager.read(filename)

        def _is_persistent_stored_callback(complete_name, success):
            if not success:
                print(f'Persistent params: failed to store {complete_name}!')
                QMessageBox.about(self, 'Warning', f'Failed to persistently store {complete_name}!')
            else:
                print(f'Persistent params: stored {complete_name}!')

        _set_param_names = []
        for param, state in parameters.items():
            if state.is_stored:
                try:
                    self.cf.param.set_value(param, state.stored_value)
                    _set_param_names.append(param)
                except Exception:
                    print(f'Failed to set {param}!')
                    QMessageBox.about(self, 'Warning', f'Failed to set {param}!')
                print(f'Set {param}!')
                self.cf.param.persistent_store(param, _is_persistent_stored_callback)

        self._update_param_io_buttons()
        dlg = QMessageBox(self)
        dlg.setWindowTitle("Info")
        _parameters_and_values = [f"{_param_name}:{parameters[_param_name].stored_value}"
                                  for _param_name in _set_param_names]
        dlg.setText('Loaded persistent parameters from file:\n' + "\n".join(_parameters_and_values))
        dlg.setIcon(QMessageBox.Icon.NoIcon)
        dlg.exec()

    def _get_persistent_state(self, complete_param_name):
        wait_for_callback_event = Event()
        state_value = None

        def state_callback(complete_name, value):
            nonlocal state_value
            state_value = value
            wait_for_callback_event.set()

        self.cf.param.persistent_get_state(complete_param_name, state_callback)
        wait_for_callback_event.wait()
        return state_value

    def _get_all_persistent_param_names(self):
        persistent_params = []
        for group_name, params in self.cf.param.toc.toc.items():
            for param_name, element in params.items():
                if element.is_persistent():
                    complete_name = group_name + '.' + param_name
                    persistent_params.append(complete_name)

        return persistent_params

    def _get_all_stored_persistent_param_names(self):
        persistent_params = self._get_all_persistent_param_names()
        stored_params = []
        for complete_name in persistent_params:
            state = self._get_persistent_state(complete_name)
            if state.is_stored:
                stored_params.append(complete_name)
        return stored_params

    def _get_all_stored_persistent_params(self):
        persistent_params = self._get_all_persistent_param_names()
        stored_params = {}
        for complete_name in persistent_params:
            state = self._get_persistent_state(complete_name)
            if state.is_stored:
                stored_params[complete_name] = state
        return stored_params

    def _dump_param_button_clicked(self):
        stored_persistent_params = self._get_all_stored_persistent_params()
        names = QFileDialog.getSaveFileName(self, 'Save file', cfclient.config_path, FILE_REGEX_YAML)
        if names[0] == '':
            return
        if not names[0].endswith(".yaml"):
            filename = names[0] + ".yaml"
        else:
            filename = names[0]

        ParamFileManager.write(filename, stored_persistent_params)
        dlg = QMessageBox(self)
        dlg.setWindowTitle('Info')
        _parameters_and_values = [f"{_param_name}: {stored_persistent_params[_param_name].stored_value}"
                                  for _param_name in stored_persistent_params.keys()]
        dlg.setText('Dumped persistent parameters to file:\n' + "\n".join(_parameters_and_values))
        dlg.setIcon(QMessageBox.Icon.NoIcon)
        dlg.exec()

    def _clear_persistent_parameter(self, complete_param_name):
        wait_for_callback_event = Event()

        def is_stored_cleared(complete_name, success):
            if success:
                print(f'Persistent params: cleared {complete_name}!')
            else:
                print(f'Persistent params: failed to clear {complete_name}!')
            wait_for_callback_event.set()

        self.cf.param.persistent_clear(complete_param_name, callback=is_stored_cleared)
        wait_for_callback_event.wait()

    def _clear_stored_persistent_params_button_clicked(self):
        dlg = QMessageBox(self)
        dlg.setWindowTitle("Clear Stored Parameters Confirmation")
        dlg.setText("Are you sure you want to clear your stored persistent parameter?")
        dlg.setStandardButtons(QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No)
        button = dlg.exec()

        if button == QMessageBox.StandardButton.Yes:
            stored_persistent_params = self._get_all_stored_persistent_param_names()
            for complete_name in stored_persistent_params:
                self._clear_persistent_parameter(complete_name)

    def _connected(self, link_uri):
        self._model.reset()
        self._model.set_toc(self.cf.param.toc.toc, self._helper.cf)
        self._model.set_enabled(True)
        self._helper.cf.param.request_update_of_all_params()
        self._is_connected = True
        self._update_param_io_buttons()

    def _disconnected(self, link_uri):
        self._is_connected = False
        self._update_param_io_buttons()
        self._model.reset()
        self._paramChanged()
        self._model.set_enabled(False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2023 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
# USA.
"""
This tab plots different logging data defined by configurations that has been
pre-configured.
"""
import logging

import cfclient
from PyQt6.QtCore import pyqtSignal
from PyQt6.QtWidgets import QMessageBox
from cfclient.ui.tab_toolbox import TabToolbox
from cflib.crazyflie.log import LogConfig
from PyQt6 import QtCore
from PyQt6 import QtGui
from PyQt6 import QtNetwork
from PyQt6 import QtWebKit
from PyQt6 import uic

__author__ = 'Bitcraze AB'
__all__ = ['GpsTab']

logger = logging.getLogger(__name__)

gps_tab_class = uic.loadUiType(cfclient.module_path +
                               "/ui/tabs/gpsTab.ui")[0]


class GpsTab(TabToolbox, gps_tab_class):
    """Tab for plotting logging data"""

    _log_data_signal = pyqtSignal(int, object, object)
    _log_error_signal = pyqtSignal(object, str)

    _disconnected_signal = pyqtSignal(str)
    _connected_signal = pyqtSignal(str)
    _console_signal = pyqtSignal(str)

    def __init__(self, helper):
        super(GpsTab, self).__init__(helper, 'GPS')
        self.setupUi(self)
        self._cf = helper.cf

        view = self.view = QtWebKit.QWebView()

        cache = QtNetwork.QNetworkDiskCache()
        cache.setCacheDirectory(cfclient.config_path + "/cache")
        view.page().networkAccessManager().setCache(cache)
        view.page().networkAccessManager()

        view.page().mainFrame().addToJavaScriptWindowObject("MainWindow", self)
        view.page().setLinkDelegationPolicy(QtWebKit.QWebPage.DelegateAllLinks)
        view.load(QtCore.QUrl(cfclient.module_path + "/resources/map.html"))
        view.loadFinished.connect(self.onLoadFinished)
        view.linkClicked.connect(QtGui.QDesktopServices.openUrl)

        self.map_layout.addWidget(view)

        self._reset_max_btn.clicked.connect(self._reset_max)

        # Connect the signals
        self._log_data_signal.connect(self._log_data_received)
        self._log_error_signal.connect(self._logging_error)
        self._connected_signal.connect(self._connected)
        self._disconnected_signal.connect(self._disconnected)

        # Connect the callbacks from the Crazyflie API
        self._helper.cf.disconnected.add_callback(
            self._disconnected_signal.emit)
        self._helper.cf.connected.add_callback(
            self._connected_signal.emit)

        self._max_speed = 0.0
        self._lat = 0
        self._long = 0

    def onLoadFinished(self):
        with open(cfclient.module_path + "/resources/map.js", 'r') as f:
            frame = self.view.page().mainFrame()
            frame.evaluateJavaScript(f.read())

    @QtCore.pyqtSlot(float, float)
    def onMapMove(self, lat, lng):
        return

    def panMap(self, lng, lat):
        frame = self.view.page().mainFrame()
        frame.evaluateJavaScript('map.panTo(L.latLng({}, {}));'.format(lat,
                                                                       lng))

    def _place_cf(self, lng, lat, acc):
        frame = self.view.page().mainFrame()
        frame.evaluateJavaScript('cf.setLatLng([{}, {}]);'.format(lat, lng))

    def _connected(self, link_uri):
        lg = LogConfig("GPS", 1000)
        lg.add_variable("gps.lat")
        lg.add_variable("gps.lon")
        lg.add_variable("gps.hAcc")
        lg.add_variable("gps.hMSL")
        lg.add_variable("gps.nsat")
        self._cf.log.add_config(lg)
        if lg.valid:
            lg.data_received_cb.add_callback(self._log_data_signal.emit)
            lg.error_cb.add_callback(self._log_error_signal.emit)
            lg.start()
        else:
            logger.warning("Could not setup logging block for GPS!")
        self._max_speed = 0.0

    def _disconnected(self, link_uri):
        """Callback for when the Crazyflie has been disconnected"""
        return

    def _logging_error(self, log_conf, msg):
        """Callback from the log layer when an error occurs"""
        QMessageBox.about(self, "Plot error", "Error when starting log config"
                          " [%s]: %s" % (log_conf.name, msg))

    def _reset_max(self):
        """Callback from reset button"""
        self._max_speed = 0.0
        self._speed_max.setText(str(self._max_speed))
        self._long.setText("")
        self._lat.setText("")
        self._height.setText("")
        self._speed.setText("")
        self._heading.setText("")
        self._accuracy.setText("")
        self._fix_type.setText("")

    def _log_data_received(self, timestamp, data, logconf):
        """Callback when the log layer receives new data"""
        long = float(data["gps.lon"]) / 10000000.0
        lat = float(data["gps.lat"]) / 10000000.0

        if self._lat != lat or self._long != long:
            self._long.setText("{:.6f}".format(long))
            self._lat.setText("{:.6f}".format(lat))
            self._nbr_locked_sats.setText(str(data["gps.nsat"]))
            self._height.setText("{:.2f}".format(float(data["gps.hMSL"])))
            self._place_cf(long, lat, 1)
            self._lat = lat
            self._long = long



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2023 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License along with
#  this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
The about dialog.
"""

import sys

import cfclient
import cflib.crtp
from PyQt6.QtCore import QT_VERSION_STR
from PyQt6.QtCore import PYQT_VERSION_STR
from PyQt6 import QtWidgets
from PyQt6 import uic
from PyQt6.QtCore import pyqtSignal
from cflib.crazyflie.mem import MemoryElement

__author__ = 'Bitcraze AB'
__all__ = ['AboutDialog']

(about_widget_class,
 about_widget_base_class) = (uic.loadUiType(cfclient.module_path +
                                            '/ui/dialogs/about.ui'))

DEBUG_INFO_FORMAT = """
<b>Cfclient</b><br>
Cfclient version: {version}<br>
System: {system}<br>
Python: {pmajor}.{pminor}.{pmicro}<br>
Qt: {qt_version}<br>
PyQt: {pyqt_version}<br>
<br>
<b>Interface status</b><br>
{interface_status}
<br>
<b>Input readers</b><br>
{input_readers}
<br>
<b>Input devices</b><br>
{input_devices}
<br>
<b>Crazyflie</b><br>
Connected: {uri}<br>
Firmware: {firmware}<br>
<br>
<b>Decks found</b><br>
{decks}
<br>
<b>Sensors found</b><br>
{imu_sensors}
<br>
<b>Sensors tests</b><br>
{imu_sensor_tests}
"""

INTERFACE_FORMAT = "{}: {}<br>"
INPUT_READER_FORMAT = "{} ({} devices connected)<br>"
DEVICE_FORMAT = "{}: ({}) {}<br>"
IMU_SENSORS_FORMAT = "{}: {}<br>"
SENSOR_TESTS_FORMAT = "{}: {}<br>"
FIRMWARE_FORMAT = "{:x}{:x} ({})"
DECK_FORMAT = "{}: rev={}, adr={}<br>"


class AboutDialog(QtWidgets.QWidget, about_widget_class):
    _disconnected_signal = pyqtSignal(str)
    _cb_deck_data_updated_signal = pyqtSignal(object)

    """Crazyflie client About box for debugging and information"""

    def __init__(self, helper, *args):
        super(AboutDialog, self).__init__(*args)
        self.setupUi(self)
        self._close_button.clicked.connect(self.close)
        self._name_label.setText(
            self._name_label.text().replace('#version#', cfclient.VERSION))

        self._interface_text = ""
        self._imu_sensors_text = ""
        self._imu_sensor_test_text = ""
        self._decks_text = ""
        self._uri = None
        self._fw_rev0 = None
        self._fw_rev1 = None
        self._fw_modified = None
        self._firmware = None

        self._helper = helper

        helper.cf.param.add_update_callback(
            group="imu_sensors", cb=self._imu_sensors_update)
        helper.cf.param.add_update_callback(
            group="imu_tests", cb=self._imu_sensor_tests_update)
        helper.cf.param.add_update_callback(
            group="firmware", cb=self._firmware_update)
        helper.cf.connected.add_callback(self._connected)

        self._disconnected_signal.connect(self._disconnected)
        helper.cf.disconnected.add_callback(self._disconnected_signal.emit)

        self._cb_deck_data_updated_signal.connect(self._deck_data_updated)

    def showEvent(self, event):
        """Event when the about box is shown"""
        self._interface_text = ""
        interface_status = cflib.crtp.get_interfaces_status()
        for key in list(interface_status.keys()):
            self._interface_text += INTERFACE_FORMAT.format(
                key, interface_status[key])

        self._device_text = ""
        devs = self._helper.inputDeviceReader.available_devices()
        for d in devs:
            self._device_text += DEVICE_FORMAT.format(
                d.reader_name, d.id, d.name)
        if len(self._device_text) == 0:
            self._device_text = "None<br>"

        self._input_readers_text = ""
        # readers = self._helper.inputDeviceReader.getAvailableDevices()
        for reader in cfclient.utils.input.inputreaders.initialized_readers:
            self._input_readers_text += INPUT_READER_FORMAT.format(
                reader.name, len(reader.devices()))
        if len(self._input_readers_text) == 0:
            self._input_readers_text = "None<br>"

        if self._uri:
            self._firmware = FIRMWARE_FORMAT.format(
                self._fw_rev0,
                self._fw_rev1,
                "MODIFIED" if self._fw_modified else "CLEAN")

            self._request_deck_data_update()

        self._update_debug_info_view()

    def _update_debug_info_view(self):
        self._debug_out.setHtml(
            DEBUG_INFO_FORMAT.format(
                version=cfclient.VERSION,
                system=sys.platform,
                pmajor=sys.version_info.major,
                pminor=sys.version_info.minor,
                pmicro=sys.version_info.micro,
                qt_version=QT_VERSION_STR,
                pyqt_version=PYQT_VERSION_STR,
                interface_status=self._interface_text,
                input_devices=self._device_text,
                input_readers=self._input_readers_text,
                uri=self._uri,
                firmware=self._firmware,
                imu_sensors=self._imu_sensors_text,
                imu_sensor_tests=self._imu_sensor_test_text,
                decks=self._decks_text))

    def _connected(self, uri):
        """Callback when Crazyflie is connected"""
        self._uri = uri

    def _firmware_update(self, name, value):
        """Callback for firmware parameters"""
        if "revision0" in name:
            self._fw_rev0 = eval(value)
        if "revision1" in name:
            self._fw_rev1 = eval(value)
        if "modified" in name:
            self._fw_modified = eval(value)

    def _imu_sensors_update(self, name, value):
        """Callback for sensor found parameters"""
        param = name[name.index('.') + 1:]
        if param not in self._imu_sensors_text:
            self._imu_sensors_text += IMU_SENSORS_FORMAT.format(
                param, eval(value))

    def _imu_sensor_tests_update(self, name, value):
        """Callback for sensor test parameters"""
        param = name[name.index('.') + 1:]
        if param not in self._imu_sensor_test_text:
            self._imu_sensor_test_text += SENSOR_TESTS_FORMAT.format(
                param, eval(value))

    def _disconnected(self, uri):
        """Callback for Crazyflie disconnected"""
        self._interface_text = ""
        self._imu_sensors_text = ""
        self._imu_sensor_test_text = ""
        self._decks_text = ""
        self._uri = None
        self._fw_rev1 = None
        self._fw_rev0 = None
        self._fw_modified = None
        self._firmware = None

    def _request_deck_data_update(self):
        self._decks_text = ""
        mems = self._helper.cf.mem.get_mems(MemoryElement.TYPE_1W)
        for mem in mems:
            mem.update(self._cb_deck_data_updated_signal.emit)

    def _deck_data_updated(self, deck_data):
        name = 'N/A'
        if "Board name" in deck_data.elements:
            name = deck_data.elements["Board name"]

        rev = 'N/A'
        if "Board revision" in deck_data.elements:
            rev = deck_data.elements["Board revision"]

        self._decks_text += DECK_FORMAT.format(name, rev, deck_data.addr)

        self._update_debug_info_view()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2023 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License along with
#  this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
Shows information from the Python logging framework
"""

import logging

from PyQt6 import uic
from PyQt6.QtCore import pyqtSignal

import cfclient
from cfclient.ui.tab_toolbox import TabToolbox

__author__ = 'Bitcraze AB'
__all__ = ['LogClientTab']

logger = logging.getLogger(__name__)

log_client_tab_class = uic.loadUiType(cfclient.module_path + "/ui/tabs/logClientTab.ui")[0]


class LogHandler(logging.StreamHandler):
    def __init__(self, signal):
        logging.StreamHandler.__init__(self)
        self._signal = signal

    def emit(self, record):
        fmt = '%(levelname)s:%(name)s:%(message)s'
        formatter = logging.Formatter(fmt)
        #
        # Calling .emit() on the signal will make the callback
        # we registered in LogClientTab run.
        #
        self._signal.emit(formatter.format(record))


class LogClientTab(TabToolbox, log_client_tab_class):
    """
    A tab for showing client logging information, such
    as USB Gamepad connections or scan feedback.
    """
    _update = pyqtSignal(str)

    def __init__(self, helper):
        super(LogClientTab, self).__init__(helper, 'Log Client')
        self.setupUi(self)

        self._update.connect(self.printText)
        self._clearButton.clicked.connect(self.clear)

        cflogger = logging.getLogger(None)
        cflogger.addHandler(LogHandler(self._update))

    def printText(self, text):
        logger.debug("[%s]", text)
        self.syslog.insertPlainText(text + '\n')

    def clear(self):
        self.syslog.clear()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2013-2014 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License along with
#  this program; if not, write to the Free Software Foundation, Inc., 51
#  Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
Implementation of a periodic timer that will call a callback every time
the timer expires once started.
"""

import logging
from threading import Thread
from cflib.utils.callbacks import Caller
import time

__author__ = 'Bitcraze AB'
__all__ = ['PeriodicTimer']

logger = logging.getLogger(__name__)


class PeriodicTimer:
    """Create a periodic timer that will periodically call a callback"""

    def __init__(self, period, callback):
        self._callbacks = Caller()
        self._callbacks.add_callback(callback)
        self._started = False
        self._period = period
        self._thread = None

    def start(self):
        """Start the timer"""
        if self._thread:
            logger.warning("Timer already started, not restarting")
            return
        self._thread = _PeriodicTimerThread(self._period, self._callbacks)
        self._thread.daemon = True
        self._thread.start()

    def stop(self):
        """Stop the timer"""
        if self._thread:
            self._thread.stop()
            self._thread = None


class _PeriodicTimerThread(Thread):

    def __init__(self, period, caller):
        super(_PeriodicTimerThread, self).__init__()
        self._period = period
        self._callbacks = caller
        self._stop = False

    def stop(self):
        self._stop = True

    def run(self):
        while not self._stop:
            time.sleep(self._period)
            if self._stop:
                break
            self._callbacks.call()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2014 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
#  USA.

"""
Leap Motion reader for controlling the Crazyflie. Note that this reader needs
the Leap Motion SDK to be manually copied. See src/leapsdk/__init__.py for
more info.
"""

try:
    import leapsdk.Leap as Leap
except Exception as e:
    raise Exception(
        "Leap Motion library probably not installed ({})".format(e))

import logging

__author__ = 'Bitcraze AB'
__all__ = ['LeapmotionReader']

logger = logging.getLogger(__name__)

MODULE_MAIN = "LeapmotionReader"
MODULE_NAME = "Leap Motion"


class LeapListener(Leap.Listener):

    def set_data_callback(self, callback):
        self._dcb = callback
        self._nbr_of_fingers = 0

    def on_init(self, controller):
        logger.info("Initialized")

    def on_connect(self, controller):
        logger.info("Connected")

    def on_disconnect(self, controller):
        # Note: not dispatched when running in a debugger.
        logger.info("Disconnected")

    def on_exit(self, controller):
        logger.info("Exited")

    def nbr_of_fingers(self):
        return self._nbr_of_fingers

    def on_frame(self, controller):
        # Get the most recent frame and report some basic information
        frame = controller.frame()
        data = {"roll": 0, "pitch": 0, "yaw": 0, "thrust": 0}
        if not frame.hands.is_empty:
            # Get the first hand
            hand = frame.hands[0]

            normal = hand.palm_normal
            direction = hand.direction
            # Pich and roll are mixed up...

            if len(hand.fingers) >= 4:
                data["roll"] = -direction.pitch * Leap.RAD_TO_DEG / 30.0
                data["pitch"] = -normal.roll * Leap.RAD_TO_DEG / 30.0
                data["yaw"] = direction.yaw * Leap.RAD_TO_DEG / 70.0
                # Use the elevation of the hand for thrust
                data["thrust"] = (hand.palm_position[1] - 80) / 150.0

            if data["thrust"] < 0.0:
                data["thrust"] = 0.0
            if data["thrust"] > 1.0:
                data["thrust"] = 1.0

        self._dcb(data)


class LeapmotionReader:
    """Used for reading data from input devices using the PyGame API."""

    def __init__(self):
        # pygame.init()
        self._ts = 0
        self._listener = LeapListener()
        self._listener.set_data_callback(self.leap_callback)
        self._controller = Leap.Controller()
        self._controller.add_listener(self._listener)
        self.name = MODULE_NAME

        self.limit_rp = True
        self.limit_thrust = True
        self.limit_yaw = True

        self.data = {"roll": 0.0, "pitch": 0.0, "yaw": 0.0,
                     "thrust": -1.0, "estop": False, "exit": False,
                     "assistedControl": False, "alt1": False, "alt2": False,
                     "pitchNeg": False, "rollNeg": False,
                     "pitchPos": False, "rollPos": False}
        logger.info("Initialized Leap")

    def open(self, deviceId):
        """
        Initialize the reading and open the device with deviceId and set the
        mapping for axis/buttons using the inputMap
        """
        return

    def leap_callback(self, data):
        for k in list(data.keys()):
            self.data[k] = data[k]

    def read(self, id):
        """Read input from the selected device."""
        return self.data

    def close(self, id):
        return

    def devices(self):
        """List all the available devices."""
        dev = []

        # According to API doc only 0 or 1 devices is supported
        # logger.info("Devs: {}".format(self._controller.is_connected))
        if self._controller.is_connected:
            dev.append({"id": 0, "name": "Leapmotion"})

        return dev



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2023 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License along with
#  this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
Container for the simple plot with functionality for data legend, saving data
and manipulating the plot.

For more advanced plotting save the data and use an external application.
"""

from PyQt6 import QtWidgets, uic

from time import time

import logging

from PyQt6.QtWidgets import QButtonGroup
from PyQt6.QtCore import *  # noqa
from PyQt6.QtWidgets import *  # noqa

import cfclient

__author__ = 'Bitcraze AB'
__all__ = ['PlotWidget']

logger = logging.getLogger(__name__)

(plot_widget_class, connect_widget_base_class) = (
    uic.loadUiType(cfclient.module_path + '/ui/widgets/plotter.ui'))

# Try the imports for PyQtGraph to see if it is installed
try:
    import pyqtgraph as pg
    from pyqtgraph import ViewBox
    import pyqtgraph.console  # noqa
    import numpy as np  # noqa

    _pyqtgraph_found = True
except Exception:
    import traceback

    logger.warning("PyQtGraph (or dependency) failed to import:\n%s",
                   traceback.format_exc())
    _pyqtgraph_found = False

# This is required to force py2exe to pull in the correct dependencies on
# Windows. But for Linux this is not required and might not be installed with
# the PyQtGraph package.
try:
    from scipy.stats import futil  # noqa
    from scipy.sparse.csgraph import _validation  # noqa
    from scipy.special import _ufuncs_cxx  # noqa
except Exception:
    pass


class PlotItemWrapper:
    """Wrapper for PlotDataItem to handle what data is shown"""

    def __init__(self, curve):
        """Initialize"""
        self.data = []
        self.ts = []
        self.curve = curve

    def add_point(self, p, ts):
        """
        Add a point to the curve.

        p - point
        ts - timestamp in ms
        """
        self.data.append(p)
        self.ts.append(ts)

    def show_data(self, start, stop):
        """
        Set what data should be shown from the curve. This is done to keep
        performance when many points have been added.
        """
        limit = min(stop, len(self.data))
        self.curve.setData(y=self.data[start:limit], x=self.ts[start:limit])
        return [self.ts[start], self.ts[limit - 1]]


class PlotWidget(QtWidgets.QWidget, plot_widget_class):
    """Wrapper widget for PyQtGraph adding some extra buttons"""

    def __init__(self, parent=None, fps=100, title="", *args):
        super(PlotWidget, self).__init__(*args)
        self.setupUi(self)

        # Limit the plot update to 10Hz
        self._ts = time()
        self._delay = 0.1

        # Check if we could import PyQtGraph, if not then stop here
        if not _pyqtgraph_found:
            self.can_enable = False
            return
        else:
            self.can_enable = True

        self._items = {}
        self._last_item = 0

        self.setSizePolicy(QtWidgets.QSizePolicy(
            QtWidgets.QSizePolicy.Policy.MinimumExpanding,
            QtWidgets.QSizePolicy.Policy.MinimumExpanding))

        self.setMinimumSize(self.minimumSizeHint())
        self.parent = parent

        pg.setConfigOption('background', 'w')
        pg.setConfigOption('foreground', 'k')
        self._plot_widget = pg.PlotWidget()
        self._plot_widget.hideButtons()
        self._plot_widget.setLabel('bottom', "Time", "ms")
        self._plot_widget.addLegend()
        self._plot_widget.getViewBox().disableAutoRange(ViewBox.XAxis)
        self._plot_widget.getViewBox().sigRangeChangedManually.connect(
            self._manual_range_change)
        self._plot_widget.getViewBox().setMouseEnabled(x=False, y=True)
        self._plot_widget.getViewBox().setMouseMode(ViewBox.PanMode)

        self.plotLayout.addWidget(self._plot_widget)

        # self.saveToFile.clicked.connect(self.saveToFileSignal)
        self._x_min = 0
        self._x_max = 500
        self._enable_auto_y.setChecked(True)
        self._enable_samples_x.setChecked(True)
        self._last_ts = None
        self._dtime = None
        self._first_ts = None

        self._x_range = (
            float(self._range_x_min.text().replace(',', '.')), float(self._range_x_max.text().replace(',', '.'))
        )
        self._nbr_samples = int(self._nbr_of_samples_x.text())
        self._nbr_of_samples_x.valueChanged.connect(self._nbr_samples_changed)
        self._nbr_seconds = int(self._nbr_of_seconds_x.text())
        self._nbr_of_seconds_x.valueChanged.connect(self._nbr_of_seconds_changed)
        self._range_x_min.valueChanged.connect(self._x_range_changed)
        self._range_x_max.valueChanged.connect(self._x_range_changed)

        self._range_y_min.valueChanged.connect(self._y_range_changed)
        self._range_y_max.valueChanged.connect(self._y_range_changed)

        self._y_btn_group = QButtonGroup()
        self._y_btn_group.addButton(self._enable_auto_y)
        self._y_btn_group.addButton(self._enable_range_y)
        self._y_btn_group.setExclusive(True)
        self._y_btn_group.buttonClicked.connect(self._y_mode_change)

        self._x_btn_group = QButtonGroup()
        self._x_btn_group.addButton(self._enable_range_x)
        self._x_btn_group.addButton(self._enable_samples_x)
        self._x_btn_group.addButton(self._enable_seconds_x)
        self._x_btn_group.setExclusive(True)
        self._x_btn_group.buttonClicked.connect(self._x_mode_change)

        self._draw_graph = True
        self._auto_redraw.stateChanged.connect(self._auto_redraw_change)

    def _auto_redraw_change(self, state):
        """Callback from the auto redraw checkbox"""
        if state == 0:
            self._draw_graph = False
        else:
            self._draw_graph = True

    def _x_mode_change(self, box):
        """Callback when user changes the X-axis mode"""
        self._nbr_of_samples_x.setEnabled(False)
        self._nbr_of_seconds_x.setEnabled(False)
        self._range_x_min.setEnabled(False)
        self._range_x_max.setEnabled(False)
        if box == self._enable_range_x:
            self._range_x_min.setEnabled(True)
            self._range_x_max.setEnabled(True)
        elif box == self._enable_samples_x:
            self._nbr_of_samples_x.setEnabled(True)
        elif box == self._enable_seconds_x:
            self._nbr_of_seconds_x.setEnabled(True)

    def _y_mode_change(self, box):
        """Callback when user changes the Y-axis mode"""
        if box == self._enable_range_y:
            self._range_y_min.setEnabled(True)
            self._range_y_max.setEnabled(True)
            y_range = (
                float(self._range_y_min.value()),
                float(self._range_y_max.value()))
            self._plot_widget.getViewBox().setRange(yRange=y_range)
        else:
            self._range_y_min.setEnabled(False)
            self._range_y_max.setEnabled(False)

        if box == self._enable_auto_y:
            self._plot_widget.getViewBox().enableAutoRange(ViewBox.YAxis)

    def _manual_range_change(self, obj):
        """
        Callback from pyqtplot when users changes the range of the plot using
        the mouse
        """
        [[x_min, x_max],
         [y_min, y_max]] = self._plot_widget.getViewBox().viewRange()
        self._range_y_min.setValue(y_min)
        self._range_y_max.setValue(y_max)
        self._range_y_min.setEnabled(True)
        self._range_y_max.setEnabled(True)
        self._enable_range_y.setChecked(True)

    def _y_range_changed(self, val):
        """Callback when user changes Y range manually"""
        _y_range = (
            float(self._range_y_min.value()),
            float(self._range_y_max.value()))
        self._plot_widget.getViewBox().setRange(yRange=_y_range, padding=0)

    def _nbr_samples_changed(self, val):
        """Callback when user changes the number of samples to be shown"""
        self._nbr_samples = val

    def _nbr_of_seconds_changed(self, val):
        """Callback when user changes the number of seconds to be shown"""
        self._nbr_seconds = val

    def _x_range_changed(self, val):
        self._range_x_min.setMaximum(self._range_x_max.value()-1)
        self._range_x_max.setMinimum(self._range_x_min.value()+1)

    def set_title(self, title):
        """
        Set the title of the plot.

        title - the new title
        """
        self._plot_widget.setTitle(title)

    def add_curve(self, title, pen='r'):
        """
        Add a new curve to the plot.

        title - the name of the data
        pen - color of curve (using r for red and so on..)
        """
        self._items[title] = PlotItemWrapper(
            self._plot_widget.plot(name=title, pen=pen))

    def add_data(self, data, ts):
        """
        Add new data to the plot.

        data - dictionary sent from logging layer containing variable/value
               pairs
        ts - timestamp of the data in ms
        """

        if self._first_ts is None:
            self._first_ts = ts

        if self._last_ts is None:
            self._dtime = 1e12
        else:
            self._dtime = ts - self._last_ts
        self._last_ts = ts

        x_min_limit = 0
        x_max_limit = 0
        # We are adding new datasets, calculate what we should show.
        if self._enable_samples_x.isChecked():
            x_min_limit = max(0, self._last_item - self._nbr_samples)
            x_max_limit = max(self._last_item, self._nbr_samples)
            self._range_x_min.setValue(int(self._first_ts + x_min_limit * self._dtime)/1000.)
            self._range_x_max.setValue(int(self._first_ts + x_max_limit * self._dtime)/1000.)
        elif self._enable_seconds_x.isChecked():
            x_min_limit = max(0, int(((self._last_ts - self._first_ts) - self._nbr_seconds * 1000.) / self._dtime))
            x_max_limit = max(0, self._last_item)
            self._range_x_min.setValue((self._first_ts + x_min_limit * self._dtime)/1000.)
            self._range_x_max.setValue((self._first_ts + x_min_limit * self._dtime)/1000. + self._nbr_seconds)
        elif self._enable_range_x.isChecked():
            x_min_limit = max(0, int((self._range_x_min.value() * 1000. - self._first_ts) / self._dtime))
            x_max_limit = max(0, int((self._range_x_max.value() * 1000. - self._first_ts) / self._dtime))

        for name in self._items:
            self._items[name].add_point(data[name], ts)
            if self._draw_graph and time() > self._ts + self._delay:
                [self._x_min, self._x_max] = self._items[name].show_data(
                    x_min_limit, x_max_limit)
        if time() > self._ts + self._delay:
            self._ts = time()
        if (self._enable_samples_x.isChecked() and self._dtime and
                self._last_item < self._nbr_samples):
            self._x_max = self._x_min + self._nbr_samples * self._dtime
        elif (self._enable_seconds_x.isChecked() and self._dtime and
                self._last_item < int(self._nbr_seconds * 1000. / self._dtime)):
            self._x_max = self._x_min + self._nbr_seconds * 1000.
        elif (self._enable_range_x.isChecked() and self._dtime) and self._last_item < x_max_limit:
            self._x_max = self._x_min + (x_max_limit - x_min_limit) * self._dtime

        self._last_item = self._last_item + 1
        self._plot_widget.getViewBox().setRange(
            xRange=(self._x_min, self._x_max))

    def removeAllDatasets(self):
        """Reset the plot by removing all the datasets"""
        for item in self._items:
            self._plot_widget.removeItem(self._items[item])

        self._clear_legend()

        self._items = {}
        self._last_item = 0
        self._last_ts = None
        self._first_ts = None
        self._dtime = None
        self._plot_widget.clear()

    def _clear_legend(self):
        legend = self._plot_widget.plotItem.legend

        while legend.layout.count() > 0:
            item = legend.items[0]
            name = item[1].text
            legend.removeItem(name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Original code from rshum19


"""
Code showing how to control the Crazyflie using the ZMQ input socket.
This code will ramp the Crazyflie motors from 25% to 45%

To work, ZMQ has to be enabled in the client configuration and the client needs
to be connected to a Crazyflie.
Search in bitcraze.io for 'ZMQ implementation of the cfclient' under header 'input devices'
for the protocol documentation.
"""

import time

try:
    import zmq
except ImportError as e:
    raise Exception("ZMQ library probably not installed ({})".format(e))

context = zmq.Context()
sender = context.socket(zmq.PUSH)
bind_addr = "tcp://127.0.0.1:{}".format(1024 + 188)
sender.connect(bind_addr)

cmdmess = {
    "version": 1,
    "ctrl": {
        "roll": 0.0,
        "pitch": 0.0,
        "yaw": 0.0,
        "thrust": 30
    }
}
print("starting to send control commands!")

# Unlocking thrust protection
cmdmess["ctrl"]["thrust"] = 0
sender.send_json(cmdmess)

for i in range(2500, 4500, 1):
    cmdmess["ctrl"]["thrust"] = i / 100.0
    sender.send_json(cmdmess)
    time.sleep(0.01)

cmdmess["ctrl"]["thrust"] = 0
sender.send_json(cmdmess)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA  02110-1301, USA.

"""
Gives access for reading and writing application configuration parameters
"""
import json
import logging
from .singleton import Singleton

import cfclient

__author__ = 'Bitcraze AB'
__all__ = ['Config']

logger = logging.getLogger(__name__)


class Config(metaclass=Singleton):
    """ Singleton class for accessing application configuration """

    def __init__(self):
        """ Initializes the singleton and reads the config files """
        self._dist_config = cfclient.module_path + "/configs/config.json"
        self._config = cfclient.config_path + "/config.json"

        [self._readonly, self._data] = self._read_distfile()

        user_config = self._read_config()
        if (user_config):
            self._data.update(user_config)

    def _read_distfile(self):
        """ Read the distribution config file containing the defaults """
        f = open(self._dist_config, 'r')
        data = json.load(f)
        f.close()
        logger.info("Dist config read from %s" % self._dist_config)

        return [data["read-only"], data["writable"]]

    def set(self, key, value):
        """ Set the value of a config parameter """
        try:
            self._data[key] = value
        except KeyError:
            raise KeyError("Could not set the parameter [%s]" % key)

    def get(self, key):
        """ Get the value of a config parameter """
        value = None
        if (key in self._data):
            value = self._data[key]
        elif (key in self._readonly):
            value = self._readonly[key]
        else:
            raise KeyError("Could not get the parameter [%s]" % key)

        if (isinstance(value, str)):
            value = str(value)

        return value

    def save_file(self):
        """ Save the user config to file """
        json_data = open(self._config, 'w')
        json_data.write(json.dumps(self._data, indent=2))
        json_data.close()
        logger.info("Config file saved to [%s]" % self._config)

    def _read_config(self):
        """ Read the user config from file """
        try:
            json_data = open(self._config)
            data = json.load(json_data)
            json_data.close()
            logger.info("Config file read from [%s]" % self._config)
        except Exception:
            return None

        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2023 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License along with
#  this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
Attitude indicator widget.
"""

import sys

from PyQt6 import QtGui
from PyQt6 import QtWidgets
from PyQt6.QtCore import Qt

__author__ = 'Bitcraze AB'
__all__ = ['AttitudeIndicator']


class AttitudeIndicator(QtWidgets.QWidget):
    """Widget for showing attitude"""

    def __init__(self):
        super(AttitudeIndicator, self).__init__()

        self.roll = 0
        self.pitch = 0
        self.hover = False
        self.hoverHeight = 0.0
        self.hoverTargetHeight = 0.0

        self.setMinimumSize(30, 30)
        # self.setMaximumSize(240,240)

    def setRoll(self, roll, repaint=True):
        self.roll = roll
        if repaint:
            self.repaint()

    def setPitch(self, pitch, repaint=True):
        self.pitch = pitch
        if repaint:
            self.repaint()

    def setHover(self, target, repaint=True):
        self.hoverTargetHeight = target
        self.hover = target > 0
        if repaint:
            self.repaint()

    def setBaro(self, height, repaint=True):
        self.hoverHeight = height
        if repaint:
            self.repaint()

    def setRollPitch(self, roll, pitch, repaint=True):
        self.roll = roll
        self.pitch = pitch
        if repaint:
            self.repaint()

    def paintEvent(self, e):
        qp = QtGui.QPainter()
        qp.begin(self)
        self.drawWidget(qp)
        qp.end()

    def drawWidget(self, qp):
        size = self.size()
        w = size.width()
        h = size.height()

        qp.translate(w / 2, h / 2)
        qp.rotate(self.roll)
        qp.translate(0, (self.pitch * h) / 50)
        qp.translate(-w / 2, -h / 2)
        qp.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing)

        font = QtGui.QFont('Serif', 7, QtGui.QFont.Weight.Light)
        qp.setFont(font)

        # Draw the blue
        qp.setPen(QtGui.QColor(0, 61, 144))
        qp.setBrush(QtGui.QColor(0, 61, 144))
        qp.drawRect(-w, int(h / 2), 3 * w, -3 * h)

        # Draw the marron
        qp.setPen(QtGui.QColor(59, 41, 39))
        qp.setBrush(QtGui.QColor(59, 41, 39))
        qp.drawRect(-w, int(h / 2), 3 * w, 3 * h)

        pen = QtGui.QPen(QtGui.QColor(255, 255, 255), 1.5, Qt.PenStyle.SolidLine)
        qp.setPen(pen)
        qp.drawLine(-w, int(h / 2), 3 * w, int(h / 2))

        # Drawing pitch lines
        for ofset in [-180, 0, 180]:
            for i in range(-900, 900, 25):
                pos = int((((i / 10.0) + 25 + ofset) * h / 50.0))
                if i % 100 == 0:
                    length = 0.35 * w
                    if i != 0:
                        if ofset == 0:
                            qp.drawText(int((w / 2) + (length / 2) + (w * 0.06)),
                                        pos, "{}".format(-i / 10))
                            qp.drawText(int((w / 2) - (length / 2) - (w * 0.08)),
                                        pos, "{}".format(-i / 10))
                        else:
                            qp.drawText(int((w / 2) + (length / 2) + (w * 0.06)),
                                        pos, "{}".format(i / 10))
                            qp.drawText(int((w / 2) - (length / 2) - (w * 0.08)),
                                        pos, "{}".format(i / 10))
                elif i % 50 == 0:
                    length = 0.2 * w
                else:
                    length = 0.1 * w

                qp.drawLine(int((w / 2) - (length / 2)), pos,
                            int((w / 2) + (length / 2)), pos)

        qp.setWorldMatrixEnabled(False)

        pen = QtGui.QPen(QtGui.QColor(0, 0, 0), 2, Qt.PenStyle.SolidLine)
        qp.setBrush(QtGui.QColor(0, 0, 0))
        qp.setPen(pen)
        qp.drawLine(0, int(h / 2), w, int(h / 2))

        # Draw Hover vs Target

        qp.setWorldMatrixEnabled(False)

        pen = QtGui.QPen(QtGui.QColor(255, 255, 255), 2, Qt.PenStyle.SolidLine)
        qp.setBrush(QtGui.QColor(255, 255, 255))
        qp.setPen(pen)
        fh = int(max(7, h / 50))
        font = QtGui.QFont('Sans', fh, QtGui.QFont.Weight.Light)
        qp.setFont(font)
        qp.resetTransform()

        qp.translate(0, h / 2)
        if not self.hover:
            # height
            qp.drawText(w - fh * 10, int(fh / 2), str(round(self.hoverHeight, 2)))

        if self.hover:
            # target height (center)
            qp.drawText(
                w - fh * 10, int(fh / 2), str(round(self.hoverTargetHeight, 2)))
            diff = round(self.hoverHeight - self.hoverTargetHeight, 2)
            pos_y = -h / 6 * diff

            # cap to +- 2.8m
            if diff < -2.8:
                pos_y = -h / 6 * -2.8
            elif diff > 2.8:
                pos_y = -h / 6 * 2.8
            else:
                pos_y = -h / 6 * diff

            # difference from target (moves up and down +- 2.8m)
            qp.drawText(int(w - fh * 3.8), int(pos_y + fh / 2), str(diff))
            # vertical line
            qp.drawLine(int(w - fh * 4.5), 0, int(w - fh * 4.5), int(pos_y))
            # left horizontal line
            qp.drawLine(int(w - fh * 4.7), 0, int(w - fh * 4.5), 0)
            # right horizontal line
            qp.drawLine(int(w - fh * 4.2), int(pos_y), int(w - fh * 4.5), int(pos_y))


if __name__ == "__main__":
    class Example(QtWidgets.QWidget):

        def __init__(self):
            super(Example, self).__init__()

            self.initUI()

        def updatePitch(self, pitch):
            self.wid.setPitch(pitch - 90)

        def updateRoll(self, roll):
            self.wid.setRoll((roll / 10.0) - 180.0)

        def updateTarget(self, target):
            self.wid.setHover(500 + target / 10.)

        def updateBaro(self, height):
            self.wid.setBaro(500 + height / 10.)

        def initUI(self):
            vbox = QtWidgets.QVBoxLayout()

            sld = QtWidgets.QSlider(Qt.Orientation.Horizontal, self)
            sld.setFocusPolicy(Qt.FocusPolicy.NoFocus)
            sld.setRange(0, 3600)
            sld.setValue(1800)
            vbox.addWidget(sld)

            self.wid = AttitudeIndicator()

            sld.valueChanged[int].connect(self.updateRoll)
            vbox.addWidget(self.wid)

            hbox = QtWidgets.QHBoxLayout()
            hbox.addLayout(vbox)

            sldPitch = QtWidgets.QSlider(Qt.Orientation.Vertical, self)
            sldPitch.setFocusPolicy(Qt.FocusPolicy.NoFocus)
            sldPitch.setRange(0, 180)
            sldPitch.setValue(90)
            sldPitch.valueChanged[int].connect(self.updatePitch)
            hbox.addWidget(sldPitch)

            sldHeight = QtWidgets.QSlider(Qt.Orientation.Vertical, self)
            sldHeight.setFocusPolicy(Qt.FocusPolicy.NoFocus)
            sldHeight.setRange(-200, 200)
            sldHeight.setValue(0)
            sldHeight.valueChanged[int].connect(self.updateBaro)

            sldT = QtWidgets.QSlider(Qt.Orientation.Vertical, self)
            sldT.setFocusPolicy(Qt.FocusPolicy.NoFocus)
            sldT.setRange(-200, 200)
            sldT.setValue(0)
            sldT.valueChanged[int].connect(self.updateTarget)

            hbox.addWidget(sldT)
            hbox.addWidget(sldHeight)

            self.setLayout(hbox)

            self.setGeometry(50, 50, 510, 510)
            self.setWindowTitle('Attitude Indicator')
            self.show()

        def changeValue(self, value):
            self.c.updateBW.emit(value)
            self.wid.repaint()

    def main():
        app = QtWidgets.QApplication(sys.argv)
        Example()
        sys.exit(app.exec())

    if __name__ == '__main__':
        main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2023 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#  02110-1301, USA.

"""
Shows data for the Loco Positioning system
"""

import logging
from enum import Enum
from collections import namedtuple

import time
from PyQt6 import uic
from PyQt6.QtCore import Qt, pyqtSignal, QTimer
from PyQt6.QtWidgets import QMessageBox
from PyQt6.QtWidgets import QLabel

import cfclient
from cfclient.ui.tab_toolbox import TabToolbox

from cflib.crazyflie.log import LogConfig
from cflib.crazyflie.mem import MemoryElement
from lpslib.lopoanchor import LoPoAnchor

from cfclient.ui.dialogs.anchor_position_dialog import AnchorPositionDialog

from vispy import scene
import numpy as np

import copy

__author__ = 'Bitcraze AB'
__all__ = ['LocoPositioningTab']

logger = logging.getLogger(__name__)

locopositioning_tab_class = uic.loadUiType(cfclient.module_path + "/ui/tabs/locopositioning_tab.ui")[0]

STYLE_RED_BACKGROUND = "background-color: lightpink;"
STYLE_GREEN_BACKGROUND = "background-color: lightgreen;"
STYLE_NO_BACKGROUND = "background-color: none;"


class Anchor:
    def __init__(self, x=0.0, y=0.0, z=0.0, distance=0.0):
        self.x = x
        self.y = y
        self.z = z
        self._is_position_valid = False
        self._is_active = False
        self.distance = distance

    def set_position(self, position):
        """Sets the position."""
        self.x = position[0]
        self.y = position[1]
        self.z = position[2]
        self._is_position_valid = True

    def get_position(self):
        """Returns the position as a vector"""
        return (self.x, self.y, self.z)

    def is_position_valid(self):
        return self._is_position_valid

    def set_is_active(self, is_active):
        self._is_active = is_active

    def is_active(self):
        return self._is_active


class AxisScaleStep:
    def __init__(self, from_view, from_axis, to_view, to_axis,
                 center_only=False):
        self.from_view = from_view.view
        self.from_axis = from_axis
        self.to_view = to_view.view
        self.to_axis = to_axis
        self.center_only = center_only


class Plot3dLps(scene.SceneCanvas):
    ANCHOR_BRUSH = np.array((0.2, 0.5, 0.2))
    ANCHOR_BRUSH_INVALID = np.array((0.8, 0.5, 0.5))
    HIGHLIGHT_ANCHOR_BRUSH = np.array((0, 1, 0))
    POSITION_BRUSH = np.array((0, 0, 1.0))

    VICINITY_DISTANCE = 2.5
    HIGHLIGHT_DISTANCE = 0.5

    LABEL_SIZE = 100
    LABEL_HIGHLIGHT_SIZE = 200

    ANCHOR_SIZE = 10
    HIGHLIGHT_SIZE = 20

    TEXT_OFFSET = np.array((0.0, 0, 0.25))

    def __init__(self):
        scene.SceneCanvas.__init__(self, keys=None)
        self.unfreeze()

        self._view = self.central_widget.add_view()
        self._view.bgcolor = '#ffffff'
        self._view.camera = scene.TurntableCamera(
            distance=10.0,
            up='+z',
            center=(0.0, 0.0, 1.0))

        self._cf = scene.visuals.Markers(
            pos=np.array([[0, 0, 0]]),
            parent=self._view.scene,
            face_color=self.POSITION_BRUSH)
        self._anchor_contexts = {}

        self.freeze()

        plane_size = 10
        scene.visuals.Plane(
            width=plane_size,
            height=plane_size,
            width_segments=plane_size,
            height_segments=plane_size,
            color=(0.5, 0.5, 0.5, 0.5),
            edge_color="gray",
            parent=self._view.scene)

        self.addArrows(1, 0.02, 0.1, 0.1, self._view.scene)

    def addArrows(self, length, width, head_length, head_width, parent):
        # The Arrow visual in vispy does not seem to work very good,
        # draw arrows using lines instead.
        w = width / 2
        hw = head_width / 2
        base_len = length - head_length

        # X-axis
        scene.visuals.LinePlot([
            [0, w, 0],
            [base_len, w, 0],
            [base_len, hw, 0],
            [length, 0, 0],
            [base_len, -hw, 0],
            [base_len, -w, 0],
            [0, -w, 0]],
            width=1.0, color='red', parent=parent, marker_size=0.0)

        # Y-axis
        scene.visuals.LinePlot([
            [w, 0, 0],
            [w, base_len, 0],
            [hw, base_len, 0],
            [0, length, 0],
            [-hw, base_len, 0],
            [-w, base_len, 0],
            [-w, 0, 0]],
            width=1.0, color='green', parent=parent, marker_size=0.0)

        # Z-axis
        scene.visuals.LinePlot([
            [0, w, 0],
            [0, w, base_len],
            [0, hw, base_len],
            [0, 0, length],
            [0, -hw, base_len],
            [0, -w, base_len],
            [0, -w, 0]],
            width=1.0, color='blue', parent=parent, marker_size=0.0)

    def update_data(self, anchors, pos, display_mode):
        self._cf.set_data(pos=np.array([pos]), face_color=self.POSITION_BRUSH)

        for id, anchor in anchors.items():
            self._update_anchor(id, anchor, display_mode)

        self._purge_anchors(anchors.keys())

    def _update_anchor(self, id, anchor, display_mode):
        if anchor.is_active():
            color = self.ANCHOR_BRUSH
        else:
            color = self.ANCHOR_BRUSH_INVALID

        size = self.ANCHOR_SIZE
        font_size = self.LABEL_SIZE
        distance = anchor.distance
        if display_mode is DisplayMode.identify_anchor:
            if distance < self.VICINITY_DISTANCE:
                amount = (distance - self.HIGHLIGHT_DISTANCE) / \
                    (self.VICINITY_DISTANCE - self.HIGHLIGHT_DISTANCE)
                color = self._mix(color, self.HIGHLIGHT_ANCHOR_BRUSH, amount)

            if distance < self.HIGHLIGHT_DISTANCE:
                color = self.HIGHLIGHT_ANCHOR_BRUSH
                size = self.HIGHLIGHT_SIZE
                font_size = self.LABEL_HIGHLIGHT_SIZE

        marker_pos = anchor.get_position()
        text_pos = self.TEXT_OFFSET + marker_pos
        if id in self._anchor_contexts:
            self._anchor_contexts[id][0].set_data(
                pos=np.array([marker_pos]),
                face_color=color,
                size=size)

            text = self._anchor_contexts[id][1]
            text.pos = text_pos
            text.font_size = font_size
        else:
            marker = scene.visuals.Markers(
                pos=np.array([marker_pos]),
                face_color=color,
                size=size,
                parent=self._view.scene)
            text = scene.visuals.Text(
                text=str(id),
                font_size=font_size,
                pos=text_pos,
                parent=self._view.scene)
            self._anchor_contexts[id] = [marker, text]

    def _purge_anchors(self, keep):
        to_remove = []
        for id, context in self._anchor_contexts.items():
            if id not in keep:
                to_remove.append(id)
                for visual in context:
                    visual.parent = None
        for id in to_remove:
            self._anchor_contexts.pop(id)

    def _mix(self, col1, col2, mix):
        return col1 * mix + col2 * (1.0 - mix)


class DisplayMode(Enum):
    identify_anchor = 1
    estimated_position = 2


Range = namedtuple('Range', ['min', 'max'])


class AnchorStateMachine:
    GET_ACTIVE = 0
    GET_IDS = 1
    GET_DATA = 2
    STEPS = [
        GET_ACTIVE,
        GET_ACTIVE,
        GET_IDS,
        GET_ACTIVE,
        GET_ACTIVE,
        GET_DATA,
        GET_ACTIVE,
        GET_ACTIVE,
    ]

    def __init__(self, mem_sub, cb_active_id_list, cb_id_list, cb_data):
        self._current_step = 0
        self._waiting_for_response = False
        self._mem = self._get_mem(mem_sub)

        self._cb_active_id_list = cb_active_id_list
        self._cb_id_list = cb_id_list
        self._cb_data = cb_data

    def poll(self):
        if not self._waiting_for_response:
            self._next_step()
            self._waiting_for_response = self._request_step()

    def _next_step(self):
        self._current_step += 1
        if self._current_step >= len(AnchorStateMachine.STEPS):
            self._current_step = 0

    def _request_step(self):
        result = True

        action = AnchorStateMachine.STEPS[self._current_step]
        if action == AnchorStateMachine.GET_ACTIVE:
            self._mem.update_active_id_list(self._cb_active_id_list_updated)
        elif action == AnchorStateMachine.GET_IDS:
            self._mem.update_id_list(self._cb_id_list_updated)
        else:
            if self._mem.nr_of_anchors > 0:
                # Only request anchor data if we actually have anchors, otherwise the callback will never be called
                self._mem.update_data(self._cb_data_updated)
            else:
                result = False

        return result

    def _get_mem(self, mem_sub):
        mem = mem_sub.get_mems(MemoryElement.TYPE_LOCO2)
        if len(mem) > 0:
            return mem[0]
        return None

    def _cb_active_id_list_updated(self, mem_data):
        self._waiting_for_response = False
        if self._cb_active_id_list:
            self._cb_active_id_list(mem_data.active_anchor_ids)

    def _cb_id_list_updated(self, mem_data):
        self._waiting_for_response = False
        if self._cb_id_list:
            self._cb_id_list(mem_data.anchor_ids)

    def _cb_data_updated(self, mem_data):
        self._waiting_for_response = False
        if self._cb_data:
            self._cb_data(mem_data.anchor_data)


class LocoPositioningTab(TabToolbox, locopositioning_tab_class):
    """Tab for plotting Loco Positioning data"""

    # Update period of log data in ms
    UPDATE_PERIOD_LOG = 100

    # Update period of anchor position data
    UPDATE_PERIOD_ANCHOR_STATE = 1000

    UPDATE_PERIOD_LOCO_MODE = 1000

    LOCO_MODE_UNKNOWN = -1
    LOCO_MODE_AUTO = 0
    LOCO_MODE_TWR = 1
    LOCO_MODE_TDOA2 = 2
    LOCO_MODE_TDOA3 = 3

    PARAM_MDOE_GR = 'loco'
    PARAM_MODE_NM = 'mode'
    PARAM_MODE = PARAM_MDOE_GR + '.' + PARAM_MODE_NM

    # Frame rate (updates per second)
    FPS = 2

    _connected_signal = pyqtSignal(str)
    _disconnected_signal = pyqtSignal(str)
    _log_error_signal = pyqtSignal(object, str)
    _anchor_range_signal = pyqtSignal(int, object, object)
    _loco_sys_signal = pyqtSignal(int, object, object)
    _cb_param_to_detect_loco_deck_signal = pyqtSignal(object, object)

    _anchor_active_id_list_updated_signal = pyqtSignal(object)
    _anchor_data_updated_signal = pyqtSignal(object)

    def __init__(self, helper):
        super(LocoPositioningTab, self).__init__(helper, 'Loco Positioning')
        self.setupUi(self)

        self._anchors = {}
        self._clear_state()
        self._refs = []

        self._display_mode = DisplayMode.estimated_position

        # Always wrap callbacks from Crazyflie API though QT Signal/Slots
        # to avoid manipulating the UI when rendering it
        self._connected_signal.connect(self._connected)
        self._disconnected_signal.connect(self._disconnected)
        self._anchor_range_signal.connect(self._anchor_range_received)
        self._loco_sys_signal.connect(self._loco_sys_received)
        self._cb_param_to_detect_loco_deck_signal.connect(
            self._cb_param_to_detect_loco_deck)

        self._anchor_active_id_list_updated_signal.connect(
            self._active_id_list_updated)
        self._anchor_data_updated_signal.connect(
            self._anchor_data_updated)

        self._id_anchor_button.toggled.connect(
            lambda enabled:
            self._do_when_checked(
                enabled,
                self._set_display_mode,
                DisplayMode.identify_anchor)
        )

        self._estimated_postion_button.toggled.connect(
            lambda enabled:
            self._do_when_checked(
                enabled,
                self._set_display_mode,
                DisplayMode.estimated_position)
        )

        self._mode_auto.toggled.connect(
            lambda enabled: self._request_mode(enabled, self.LOCO_MODE_AUTO)
        )

        self._mode_twr.toggled.connect(
            lambda enabled: self._request_mode(enabled, self.LOCO_MODE_TWR)
        )

        self._mode_tdoa2.toggled.connect(
            lambda enabled: self._request_mode(enabled, self.LOCO_MODE_TDOA2)
        )

        self._mode_tdoa3.toggled.connect(
            lambda enabled: self._request_mode(enabled, self.LOCO_MODE_TDOA3)
        )

        self._enable_mode_buttons(False)

        self._switch_mode_to_twr_button.setEnabled(False)
        self._switch_mode_to_tdoa2_button.setEnabled(False)
        self._switch_mode_to_tdoa3_button.setEnabled(False)

        self._switch_mode_to_twr_button.clicked.connect(
            lambda enabled:
            self._send_anchor_mode(self.LOCO_MODE_TWR)
        )
        self._switch_mode_to_tdoa2_button.clicked.connect(
            lambda enabled:
            self._send_anchor_mode(self.LOCO_MODE_TDOA2)
        )
        self._switch_mode_to_tdoa3_button.clicked.connect(
            lambda enabled:
            self._send_anchor_mode(self.LOCO_MODE_TDOA3)
        )

        self._clear_anchors_button.clicked.connect(self._clear_anchors)

        self._configure_anchor_positions_button.clicked.connect(
            self._show_anchor_postion_dialog)

        # Connect the Crazyflie API callbacks to the signals
        self._helper.cf.connected.add_callback(
            self._connected_signal.emit)

        self._helper.cf.disconnected.add_callback(
            self._disconnected_signal.emit)

        self._set_up_plots()

        self.is_loco_deck_active = False

        self._graph_timer = QTimer()
        self._graph_timer.setInterval(int(1000 / self.FPS))
        self._graph_timer.timeout.connect(self._update_graphics)
        self._graph_timer.start()

        self._anchor_state_timer = QTimer()
        self._anchor_state_timer.setInterval(self.UPDATE_PERIOD_ANCHOR_STATE)
        self._anchor_state_timer.timeout.connect(self._poll_anchor_state)
        self._anchor_state_machine = None

        self._update_position_label(self._helper.pose_logger.position)

        self._lps_state = self.LOCO_MODE_UNKNOWN
        self._update_lps_state(self.LOCO_MODE_UNKNOWN)

        self._anchor_position_dialog = AnchorPositionDialog(self, helper)
        self._configure_anchor_positions_button.setEnabled(False)

    def _do_when_checked(self, enabled, fkn, arg):
        if enabled:
            fkn(arg)

    def _set_up_plots(self):
        self._plot_3d = Plot3dLps()
        self._plot_layout.addWidget(self._plot_3d.native)

    def _set_display_mode(self, display_mode):
        self._display_mode = display_mode

    def _send_anchor_mode(self, mode):
        lopo = LoPoAnchor(self._helper.cf)

        mode_translation = {
            self.LOCO_MODE_TWR: lopo.MODE_TWR,
            self.LOCO_MODE_TDOA2: lopo.MODE_TDOA,
            self.LOCO_MODE_TDOA3: lopo.MODE_TDOA3,
        }

        # Set the mode from the last to the first anchor
        # In TDoA 2 mode this ensures that the master anchor is set last
        # Note: We only switch mode of anchor 0 - 7 since this is what is
        # supported in TWR and TDoA 2
        for j in range(5):
            for i in reversed(range(8)):
                lopo.set_mode(i, mode_translation[mode])

    def _clear_state(self):
        self._clear_anchors()
        self._update_ranging_status_indicators()
        self._id_anchor_button.setEnabled(True)

    def _clear_anchors(self):
        self._anchors = {}

    def _connected(self, link_uri):
        """Callback when the Crazyflie has been connected"""
        logger.debug("Crazyflie connected to {}".format(link_uri))
        self._request_param_to_detect_loco_deck()

    def _request_param_to_detect_loco_deck(self):
        """Send a parameter request to detect if the Loco deck is installed"""
        group = 'deck'

        def register(group, param):
            if self._is_in_param_toc(group, param):
                logger.debug("Requesting loco deck parameter")
                self._helper.cf.param.add_update_callback(group=group,
                                                          name=param,
                                                          cb=self._cb_param_to_detect_loco_deck_signal.emit)

        register(group, 'bcLoco')
        register(group, 'bcDWM1000')  # For backwards compatibility

    def _cb_param_to_detect_loco_deck(self, name, value):
        """Callback from the parameter sub system when the Loco deck detection
        parameter has been updated"""
        if value == '1':
            logger.debug("Loco deck installed, enabling LPS tab")
            self._loco_deck_detected()
        else:
            logger.debug("No Loco deck installed")

    def _loco_deck_detected(self):
        """Called when the loco deck has been detected. Enables the tab,
        starts logging and polling of the memory sub system as well as starts
        timers for updating graphics"""
        if not self.is_loco_deck_active:
            self.is_loco_deck_active = True
            try:
                self._register_logblock(
                    "LoPoTab0",
                    [
                        ("ranging", "distance0", "float"),
                        ("ranging", "distance1", "float"),
                        ("ranging", "distance2", "float"),
                        ("ranging", "distance3", "float"),
                    ],
                    self._anchor_range_signal.emit,
                    self._log_error_signal.emit)

                self._register_logblock(
                    "LoPoTab1",
                    [
                        ("ranging", "distance4", "float"),
                        ("ranging", "distance5", "float"),
                        ("ranging", "distance6", "float"),
                        ("ranging", "distance7", "float"),
                    ],
                    self._anchor_range_signal.emit,
                    self._log_error_signal.emit)

                self._register_logblock(
                    "LoPoSys",
                    [
                        ("loco", "mode", "uint8_t")
                    ],
                    self._loco_sys_signal.emit,
                    self._log_error_signal.emit,
                    update_period=self.UPDATE_PERIOD_LOCO_MODE)
            except KeyError as e:
                logger.warning(str(e))
            except AttributeError as e:
                logger.warning(str(e))

            self._start_polling_anchor_pos(self._helper.cf)
            self._enable_mode_buttons(True)
            self._configure_anchor_positions_button.setEnabled(True)

            self._helper.cf.param.add_update_callback(
                group=self.PARAM_MDOE_GR,
                name=self.PARAM_MODE_NM,
                cb=self._loco_mode_updated)

            if self.PARAM_MDOE_GR in self._helper.cf.param.values:
                if self.PARAM_MODE_NM in \
                        self._helper.cf.param.values[self.PARAM_MDOE_GR]:
                    self._loco_mode_updated(
                        self.PARAM_MODE,
                        self._helper.cf.param.values[self.PARAM_MDOE_GR][
                            self.PARAM_MODE_NM])

    def _disconnected(self, link_uri):
        """Callback for when the Crazyflie has been disconnected"""
        logger.debug("Crazyflie disconnected from {}".format(link_uri))
        self._stop_polling_anchor_pos()
        self._clear_state()
        self._update_graphics()
        self.is_loco_deck_active = False
        self._update_lps_state(self.LOCO_MODE_UNKNOWN)
        self._enable_mode_buttons(False)
        self._loco_mode_updated('', self.LOCO_MODE_UNKNOWN)
        self._configure_anchor_positions_button.setEnabled(False)
        self._anchor_position_dialog.close()

    def _register_logblock(self, logblock_name, variables, data_cb, error_cb,
                           update_period=UPDATE_PERIOD_LOG):
        """Register log data to listen for. One logblock can contain a limited
        number of parameters (6 for floats)."""
        lg = LogConfig(logblock_name, update_period)
        for variable in variables:
            if self._is_in_log_toc(variable):
                lg.add_variable('{}.{}'.format(variable[0], variable[1]),
                                variable[2])

        self._helper.cf.log.add_config(lg)
        lg.data_received_cb.add_callback(data_cb)
        lg.error_cb.add_callback(error_cb)
        lg.start()
        return lg

    def _is_in_log_toc(self, variable):
        toc = self._helper.cf.log.toc
        group = variable[0]
        param = variable[1]
        return group in toc.toc and param in toc.toc[group]

    def _is_in_param_toc(self, group, param):
        toc = self._helper.cf.param.toc
        return bool(group in toc.toc and param in toc.toc[group])

    def _anchor_range_received(self, timestamp, data, logconf):
        """Callback from the logging system when a range is updated."""
        for name, value in data.items():
            valid, anchor_number = self._parse_range_param_name(name)
            # Only set distance on anchors that we have seen through other
            # messages to avoid creating anchor 0-7 even if they do not exist
            # in a TDoA3 set up for instance
            if self._anchor_exists(anchor_number):
                if valid:
                    anchor = self._get_create_anchor(anchor_number)
                    anchor.distance = float(value)

    def _loco_sys_received(self, timestamp, data, logconf):
        """Callback from the logging system when the loco pos sys config
        is updated."""
        if self.PARAM_MODE in data:
            lps_state = data[self.PARAM_MODE]
            if lps_state == self.LOCO_MODE_TDOA2:
                if self._id_anchor_button.isEnabled():
                    if self._id_anchor_button.isChecked():
                        self._estimated_postion_button.setChecked(True)
                    self._id_anchor_button.setEnabled(False)
            else:
                if not self._id_anchor_button.isEnabled():
                    self._id_anchor_button.setEnabled(True)
            self._update_lps_state(lps_state)

    def _update_ranging_status_indicators(self):
        container = self._anchor_stats_container

        ids = sorted(self._anchors.keys())

        # Update existing labels or add new if needed
        count = 0
        for id in ids:
            col = count % 8
            row = int(count / 8)

            if count < container.count():
                label = container.itemAtPosition(row, col).widget()
            else:
                label = QLabel()
                label.setMinimumSize(30, 0)
                label.setProperty('frameShape', 'QFrame::Box')
                label.setAlignment(Qt.AlignmentFlag.AlignCenter)
                container.addWidget(label, row, col)

            label.setText(str(id))

            if self._anchors[id].is_active():
                label.setStyleSheet(STYLE_GREEN_BACKGROUND)
            else:
                label.setStyleSheet(STYLE_RED_BACKGROUND)

            count += 1

        # Remove labels if there are too many
        for i in range(count, container.count()):
            col = i % 8
            row = int(i / 8)

            label = container.itemAtPosition(row, col).widget()
            label.deleteLater()

    def _logging_error(self, log_conf, msg):
        """Callback from the log layer when an error occurs"""
        QMessageBox.about(self, "LocoPositioningTab error",
                          "Error when using log config",
                          " [{0}]: {1}".format(log_conf.name, msg))

    def _start_polling_anchor_pos(self, crazyflie):
        """Set up a timer to poll anchor positions from the memory sub
        system"""
        if not self._anchor_state_machine:
            self._anchor_state_machine = AnchorStateMachine(
                crazyflie.mem,
                self._anchor_active_id_list_updated_signal.emit,
                None,
                self._anchor_data_updated_signal.emit
            )
        self._anchor_state_timer.start()

    def _stop_polling_anchor_pos(self):
        self._anchor_state_timer.stop()
        self._anchor_state_machine = None

    def _poll_anchor_state(self):
        if self._anchor_state_machine:
            self._anchor_state_machine.poll()

    def _active_id_list_updated(self, anchor_list):
        """Callback from the anchor state machine when we get a list of active
        anchors"""
        for id, anchor_data in self._anchors.items():
            anchor_data.set_is_active(False)

        for id in anchor_list:
            anchor_data = self._get_create_anchor(id)
            anchor_data.set_is_active(True)

        self._update_ranging_status_indicators()

    def _anchor_data_updated(self, position_dict):
        """Callback from the anchor state machine when the anchor positions
         are updated"""
        for id, anchor_data in position_dict.items():
            anchor = self._get_create_anchor(id)
            if anchor_data.is_valid:
                anchor.set_position(anchor_data.position)

        self._update_positions_in_config_dialog()

    def _parse_range_param_name(self, name):
        """Parse a parameter name for a ranging distance and return the number
           of the anchor. The name is on the format 'ranging.distance4' """
        valid = False
        anchor = 0
        if name.startswith('ranging.distance'):
            anchor = int(name[-1])
            valid = True
        return (valid, anchor)

    def _parse_position_param_name(self, name):
        """Parse a parameter name for a position and return the
           axis (0=x, 1=y, 2=z).
           The param name is on the format 'kalman.stateY' """
        valid = False
        axis = 0
        if name.startswith('kalman.state'):
            axis = {'X': 0, 'Y': 1, 'Z': 2}[name[-1]]
            valid = True
        return (valid, axis)

    def _get_create_anchor(self, anchor_number):
        if anchor_number not in self._anchors:
            self._anchors[anchor_number] = Anchor()
        return self._anchors[anchor_number]

    def _anchor_exists(self, anchor_number):
        return anchor_number in self._anchors

    def _update_graphics(self):
        if self.is_visible() and self.is_loco_deck_active:
            anchors = copy.deepcopy(self._anchors)
            self._plot_3d.update_data(
                anchors,
                self._helper.pose_logger.position,
                self._display_mode)
            self._update_position_label(self._helper.pose_logger.position)

    def _update_position_label(self, position):
        if len(position) == 3:
            coordinate = "({:0.2f}, {:0.2f}, {:0.2f})".format(
                position[0], position[1], position[2])
        else:
            coordinate = '(0.00, 0.00, 0.00)'

        self._status_position.setText(coordinate)

    def _update_lps_state(self, state):
        if state != self._lps_state:
            self._update_lps_state_indicator(self._state_twr,
                                             state == self.LOCO_MODE_TWR)
            self._update_lps_state_indicator(self._state_tdoa2,
                                             state == self.LOCO_MODE_TDOA2)
            self._update_lps_state_indicator(self._state_tdoa3,
                                             state == self.LOCO_MODE_TDOA3)
        self._lps_state = state

    def _update_lps_state_indicator(self, element, active):
        if active:
            element.setStyleSheet(STYLE_GREEN_BACKGROUND)
        else:
            element.setStyleSheet(STYLE_NO_BACKGROUND)

    def _enable_mode_buttons(self, enabled):
        self._mode_auto.setEnabled(enabled)
        self._mode_twr.setEnabled(enabled)
        self._mode_tdoa2.setEnabled(enabled)
        self._mode_tdoa3.setEnabled(enabled)

    def _request_mode(self, enabled, mode):
        if enabled:
            self._helper.cf.param.set_value(self.PARAM_MODE, str(mode))

            if mode == self.LOCO_MODE_TWR:
                self._switch_mode_to_twr_button.setEnabled(False)
                self._switch_mode_to_tdoa2_button.setEnabled(True)
                self._switch_mode_to_tdoa3_button.setEnabled(True)
            elif mode == self.LOCO_MODE_TDOA2:
                self._switch_mode_to_twr_button.setEnabled(True)
                self._switch_mode_to_tdoa2_button.setEnabled(False)
                self._switch_mode_to_tdoa3_button.setEnabled(True)
            elif mode == self.LOCO_MODE_TDOA3:
                self._switch_mode_to_twr_button.setEnabled(True)
                self._switch_mode_to_tdoa2_button.setEnabled(True)
                self._switch_mode_to_tdoa3_button.setEnabled(False)
            else:
                self._switch_mode_to_twr_button.setEnabled(False)
                self._switch_mode_to_tdoa2_button.setEnabled(False)
                self._switch_mode_to_tdoa3_button.setEnabled(False)

    def _loco_mode_updated(self, name, value):
        mode = int(value)
        if mode == self.LOCO_MODE_AUTO:
            if not self._mode_auto.isChecked():
                self._mode_auto.setChecked(True)
        elif mode == self.LOCO_MODE_TWR:
            if not self._mode_twr.isChecked():
                self._mode_twr.setChecked(True)
        elif mode == self.LOCO_MODE_TDOA2:
            if not self._mode_tdoa2.isChecked():
                self._mode_tdoa2.setChecked(True)
        elif mode == self.LOCO_MODE_TDOA3:
            if not self._mode_tdoa3.isChecked():
                self._mode_tdoa3.setChecked(True)
        else:
            self._mode_auto.setChecked(False)
            self._mode_twr.setChecked(False)
            self._mode_tdoa2.setChecked(False)
            self._mode_tdoa3.setChecked(False)

    def _show_anchor_postion_dialog(self):
        self._anchor_position_dialog.show()

    def _update_positions_in_config_dialog(self):
        positions = {}

        for id, anchor in self._anchors.items():
            if anchor.is_position_valid():
                positions[id] = (anchor.x, anchor.y, anchor.z)

        self._anchor_position_dialog.anchor_postions_updated(positions)

    def write_positions_to_anchors(self, anchor_positions):
        lopo = LoPoAnchor(self._helper.cf)

        for _ in range(3):
            for id, position in anchor_positions.items():
                lopo.set_position(id, position)
            time.sleep(0.2)



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2023 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License along with
#  this program; if not, write to the Free Software Foundation, Inc., 51
#  Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Initialization of the PC Client GUI."""

import platform
import sys
import os
import argparse
import datetime
import signal

import logging

import cfclient

__author__ = 'Bitcraze AB'
__all__ = []


def handle_sigint(app):
    logging.info('SIGINT received, exiting ...')
    if app:
        app.closeAllWindows()
    else:
        sys.exit(0)


def main():
    """
    Check starting conditions and start GUI.

    First, check command line arguments and start loggers. Set log levels. Try
    all imports and exit verbosely if a library is not found. Disable outputs
    to stdout and start the GUI.
    """
    app = None

    # Connect ctrl-c (SIGINT) signal
    signal.signal(signal.SIGINT, lambda sig, frame: handle_sigint(app))

    # Allows frozen mac build to load libraries from app bundle
    if getattr(sys, 'frozen', False) and platform.system() == 'Darwin':
        os.environ['DYLD_FALLBACK_LIBRARY_PATH'] = os.path.dirname(
            sys.executable)

    # Set ERROR level for PyQt logger
    qtlogger = logging.getLogger('PyQt6')
    qtlogger.setLevel(logging.ERROR)

    parser = argparse.ArgumentParser(
        description="cfclient - Crazyflie graphical control client")
    parser.add_argument('--debug', '-d', nargs=1, default='info', type=str,
                        help="set debug level "
                             "[minimal, info, debug, debugfile]")
    parser.add_argument('--check-imports', type=bool, default=False,
                        const=True, nargs="?",
                        help="Check python imports and exit successfully" +
                        " (intended for CI)")
    args = parser.parse_args()
    debug = args.debug

    cflogger = logging.getLogger('')

    # Set correct logging fuctionality according to commandline
    if ("debugfile" in debug):
        logging.basicConfig(level=logging.DEBUG)
        # Add extra format options for file logger (thread and time)
        formatter = logging.Formatter('%(asctime)s:%(threadName)s:%(name)'
                                      's:%(levelname)s:%(message)s')
        filename = "debug-%s.log" % datetime.datetime.now()
        filehandler = logging.FileHandler(filename)
        filehandler.setLevel(logging.DEBUG)
        filehandler.setFormatter(formatter)
        cflogger.addHandler(filehandler)
    elif ("debug" in debug):
        logging.basicConfig(level=logging.DEBUG)
    elif ("minimal" in debug):
        logging.basicConfig(level=logging.WARNING)
    elif ("info" in debug):
        logging.basicConfig(level=logging.INFO)

    logger = logging.getLogger(__name__)

    logger.debug("Using config path {}".format(cfclient.config_path))
    logger.debug("sys.path={}".format(sys.path))

    # Try all the imports used in the project here to control what happens....
    try:
        import usb  # noqa
    except ImportError:
        logger.critical("No pyusb installation found, exiting!")
        sys.exit(1)

    if not sys.platform.startswith('linux'):
        try:
            import sdl2  # noqa
        except ImportError:
            logger.critical("No pysdl2 installation found, exiting!")
            sys.exit(1)

    try:
        import PyQt6  # noqa
    except ImportError:
        logger.critical("No PyQT6 installation found, exiting!")
        sys.exit(1)

    # Disable printouts from STL
    if os.name == 'posix':
        stdout = os.dup(1)
        os.dup2(os.open('/dev/null', os.O_WRONLY), 1)
        sys.stdout = os.fdopen(stdout, 'w')
        logger.info("Disabling STL printouts")

    if os.name == 'nt':
        stdout = os.dup(1)
        os.dup2(os.open('NUL', os.O_WRONLY), 1)
        sys.stdout = os.fdopen(stdout, 'w')
        logger.info("Disabling STL printouts")

    if sys.platform == 'darwin':
        try:
            import Foundation
            bundle = Foundation.NSBundle.mainBundle()
            if bundle:
                info = (bundle.localizedInfoDictionary() or
                        bundle.infoDictionary())
                if info:
                    info['CFBundleName'] = 'Crazyflie'
        except ImportError:
            logger.info("Foundation not found. Menu will show python as "
                        "application name")

    if args.check_imports:
        logger.info("All imports successful!")
        sys.exit(0)

    # Start up the main user-interface
    from .ui.main import MainUI
    from PyQt6.QtWidgets import QApplication
    from PyQt6.QtGui import QIcon

    if os.name == 'posix':
        logger.info('If startup fails because of "xcb", install dependency with `sudo apt install libxcb-xinerama0`.')

    app = QApplication(sys.argv)
    app.setStyle("Fusion")
    from cfclient.utils.ui import UiUtils

    app.setWindowIcon(QIcon(cfclient.module_path + "/ui/icons/icon-256.png"))
    app.setApplicationName("Crazyflie client")
    # Make sure the right icon is set in Windows 7+ taskbar
    if os.name == 'nt':
        import ctypes

        try:
            myappid = 'mycompany.myproduct.subproduct.version'
            ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(
                myappid)
        except Exception:
            pass

    main_window = MainUI()
    app.setFont(UiUtils.FONT)
    main_window.show()
    main_window.set_default_theme()
    sys.exit(app.exec())


if __name__ == "__main__":
    main()
